#!/usr/bin/python3
# encoding=utf-8
#
# Copyright © 2015 Simon McVittie <smcv@debian.org>
# Copyright © 2023 Sébastien Noel <sebastien@twolife.be>
# SPDX-License-Identifier: GPL-2.0-or-later

from __future__ import annotations

import glob
import logging
import os
import subprocess
from typing import (TYPE_CHECKING)
from zipfile import ZipFile

from ..build import (FillResult, PackagingTask)
from ..game import (GameData)
from ..packaging import (get_native_packaging_system)
from ..util import (mkdir_p)

if TYPE_CHECKING:
    from typing import (Unpack)
    from ..data import (Package)
    from ..packaging import (PerPackageState, PackagingTaskArgs)

logger = logging.getLogger(__name__)


class Doom3GameData(GameData):
    def construct_task(self, **kwargs: Unpack[PackagingTaskArgs]) -> Doom3Task:
        return Doom3Task(self, **kwargs)


class Doom3Task(PackagingTask):
    def fill_gaps(
        self,
        package: Package | None,
        download: bool = False,
        log: bool = True,
        recheck: bool = False,
        requested: bool = False,
    ) -> FillResult:
        assert package is not None
        if package.name in (
            'doom3-classic-data',
            'doom3-the-lost-mission-data',
        ) and (
            self.packaging.get_architecture()
            != self.builder_packaging.get_architecture()
        ):
            logger.warning('Cross-compiling %s not supported', package.name)
            return FillResult.IMPOSSIBLE

        return super().fill_gaps(
            package,
            download=download,
            log=log,
            recheck=recheck,
            requested=requested,
        )

    def fill_extra_files(
        self,
        per_package_state: PerPackageState
    ) -> None:
        super().fill_extra_files(per_package_state)
        package = per_package_state.package
        destdir = per_package_state.destdir

        if package.name not in (
            'doom3-classic-data',
            'doom3-the-lost-mission-data',
        ):
            return

        libname = {
            'doom3-classic-data': 'cdoom',
            'doom3-the-lost-mission-data': 'd3le',
        }[package.name]

        # Build & install the mod lib
        packaging = get_native_packaging_system()
        assets_dir = packaging.substitute(package.install_to, package.name)
        lib_dir = packaging.get_libdir(package.architecture)
        install_all_dir = os.path.join(destdir, assets_dir[1:], libname)
        install_arch_dir = os.path.join(destdir, lib_dir[1:], 'dhewm3')
        unpackdir = os.path.join(
            self.get_workdir(), 'tmp', package.name + '.build.d',
        )

        zips = list(glob.glob(os.path.join(install_all_dir, '*.zip')))
        assert len(zips) == 1, zips
        with ZipFile(zips[0]) as zObject:
            zObject.extractall(path=unpackdir)

        expect_dir_list = list(glob.glob(os.path.join(unpackdir, '*')))
        assert len(expect_dir_list) == 1, expect_dir_list
        expect_dir = os.path.basename(expect_dir_list[0])

        sourcedir = os.path.join(unpackdir, expect_dir)
        builddir = os.path.join(unpackdir, expect_dir, 'build')
        os.mkdir(builddir)
        mkdir_p(install_arch_dir)

        subprocess.check_call([
            'cmake', '-B', builddir, '-S', sourcedir,
        ])
        subprocess.check_call([
            'make', '-C', builddir, '-s', '-j5',
        ])
        subprocess.check_call([
            'install', '-s', '-m644',
            os.path.join(builddir, libname + '.so'),
            os.path.join(install_arch_dir, libname + '.so'),
        ])

        # Desktop file
        appdir = 'usr/share/applications'
        symlinkdir = '$assets/game-data-packager-runtime'
        symlinksrc = os.path.join(symlinkdir, 'dhewm3-%s.desktop' % libname)
        symlinkdest = os.path.join(appdir, 'dhewm3-%s.desktop' % libname)
        package.symlinks[symlinkdest] = symlinksrc


GAME_DATA_SUBCLASS = Doom3GameData



#!/usr/bin/python3
# encoding=utf-8
#
# Copyright © 2018 Simon McVittie <smcv@debian.org>
# SPDX-License-Identifier: GPL-2.0-or-later

from __future__ import annotations

import re
import subprocess
from collections.abc import Collection
from types import TracebackType

from . import (SimpleUnpackable)


class InnoSetup(SimpleUnpackable):
    """Object representing an InnoSetup installer.
    """

    def __init__(
        self,
        path: str | bytes,
        verbose: bool = False,
        language: str | None = None,
    ) -> None:
        """Constructor.

        path may be a string or bytes object.
        """
        self.path = path
        self.verbose = verbose
        self.language = language

    def __enter__(self) -> InnoSetup:
        return self

    def __exit__(
        self,
        _et: type[BaseException] | None = None,
        _ev: BaseException | None = None,
        _tb: TracebackType | None = None
    ) -> None:
        pass

    def _extactall_argv(
        self,
        dest_path: str,
    ) -> list[str]:
        assert type(self.path) is str  # this contradicts  __init__ #L46
        return [
            'innoextract',
            '-T', 'local',
            '-d', dest_path,
            self.path,
        ]

    def _extractall_for_template(
        self,
        path: str
    ) -> str:
        argv = self._extactall_argv(path)
        argv.append('--collisions=rename')

        res = subprocess.check_output(
            argv,
            stderr=subprocess.DEVNULL,
            text=True,
        )
        assert type(res) is str
        return res

    def extractall(
        self,
        path: str,
        members: Collection[str] | None = None,
    ) -> None:
        argv = self._extactall_argv(path)

        if members is not None:
            for member in members:
                argv.append('-I')
                argv.append(member)

        if not self.verbose:
            argv.append('--silent')
            argv.append('--progress')

        if self.language is not None:
            argv.append('--language')
            argv.append(self.language)

        subprocess.check_call(argv)

    def printdir(self) -> None:
        subprocess.check_call([
            'innoextract',
            '--default-language', (self.language or 'english'),
            '-T', 'local',
            '--list',
            self.path,
        ])

    def search_for_gog_icon(self) -> str | None:
        result = subprocess.run([
            'innoextract',
            '--default-language', (self.language or 'english'),
            '-T', 'local',
            '--list',
            self.path,
        ], capture_output=True, check=True)
        match = re.search(
                    r' "([0-9A-Za-z/]*goggame-\d*.ico)" ',
                    result.stdout.decode()
                )
        if match:
            return match.group(1)
        return None

    @property
    def format(self) -> str:
        return 'innoextract'


if __name__ == '__main__':
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--output', '-o', help='extract to OUTPUT', default=None)
    parser.add_argument(
        '--verbose', '-v', help='Be verbose', action='store_true')
    parser.add_argument('setup_exe')
    args = parser.parse_args()

    setup = InnoSetup(args.setup_exe, verbose=args.verbose)

    if args.output:
        setup.extractall(args.output)
    else:
        setup.printdir()



#!/usr/bin/python3
# encoding=utf-8
#
# Copyright © 2015-2016 Alexandre Detiste <alexandre@detiste.be>
# Copyright © 2015 Simon McVittie <smcv@debian.org>
# Copyright © 2022 Sébastien Noel <sebastien@twolife.be>
# SPDX-License-Identifier: GPL-2.0-or-later

import os
import subprocess
from shutil import which
from typing import Any


CACODEMON = '/usr/share/pixmaps/doom2-masterlevels.png'

if os.path.isdir('/usr/share/doom'):
    DIR = '/usr/share/doom'
else:
    DIR = '/usr/share/games/doom'

if os.path.isfile('/etc/redhat-release'):
    depedencies = (
        'gtk4, python3-gobject-base and gobject-introspection\n'
        '(already pulled-in by this .rpm)'
    )
    command = 'dnf install prboom-plus'
else:
    depedencies = 'python3-gi and gir1.2-gtk-4.0'
    command = 'apt-get install doom-engine python3-gi gir1.2-gtk-4.0'

requirements = '''
--------------------------------------------------------

You need those to make use this launcher:
* the .wad files from DOOM 2 Master Levels
* some Doom game engine
* ''' + depedencies + '''

The free parts can be obtained this way:
  ''' + command + '''

The .wad files can be for example bought on Steam:
https://store.steampowered.com/app/9160/ or found
on "Doom 3: Resurrection of Evil" Xbox game disc.

The data then need to be put at the right location.
You can use game-data-packager(6) to automate this.

It will also automatically pick up the data downloaded
by a windows Steam instance running through Wine.
'''

# ValueError: Namespace Gtk not available
try:
    import gi
    gi.require_version('Gtk', '4.0')
    from gi.repository import GObject, Gio, Gtk, Gdk
except (ImportError, ValueError):
    message = 'Python3 Gtk+ libraries not found!\n' + requirements
    if which('zenity'):
        subprocess.call([
            'zenity', '--error', '--title=Doom 2 Master Levels', '--text',
            message,
        ])
    elif which('yad'):
        subprocess.call([
            'yad', '--error', '--title=Doom 2 Master Levels', '--text',
            message,
        ])
    elif which('kdialog'):
        subprocess.call([
            'kdialog', '--error', message, '--title=Doom 2 Master Levels',
        ])
    elif which('xmessage'):
        subprocess.call(['xmessage', '-center', message])
    exit(message)


setup_tools = {
    'chocolate-doom': 'chocolate-doom-setup',
    'crispy-doom': 'crispy-setup',
    'woof': 'woof-setup',
}

# wad : (warp, longname,  'https://doomwiki.org/wiki/' + url)
levels = {
    'attack.wad': (
        1, 'Attack', 'MAP01:_Attack_(Master_Levels)',
    ),
    'blacktwr.wad': (
        25, 'Black Tower', 'MAP25:_Black_Tower_(Master_Levels)',
    ),
    'bloodsea.wad': (
        7, 'Bloodsea Keep', 'MAP07:_Bloodsea_Keep_(Master_Levels)',
    ),
    'canyon.wad': (
        1, 'Canyon', 'MAP01:_Canyon_(Master_Levels)',
    ),
    'catwalk.wad': (
        1, 'The Catwalk', 'MAP01:_The_Catwalk_(Master_Levels)',
    ),
    'combine.wad': (
        1, 'The Combine', 'MAP01:_The_Combine_(Master_Levels)',
    ),
    'fistula.wad': (
        1, 'The Fistula', 'MAP01:_The_Fistula_(Master_Levels)',
    ),
    'garrison.wad': (
        1, 'The Garrison', 'MAP01:_The_Garrison_(Master_Levels)',
    ),
    'geryon.wad': (
        8, 'Geryon: 6th Canto of Inferno', 'MAP08:_Geryon_(Master_Levels)',
    ),
    'mephisto.wad': (
        7,
        "Mephisto's Maosoleum",
        "MAP07:_Mephisto%27s_Maosoleum_(Master_Levels)",
    ),
    'manor.wad': (
        1, 'Titan Manor', 'MAP01:_Titan_Manor_(Master_Levels)',
    ),
    'minos.wad': (
        5,
        "Minos' Judgement: 4th Canto of Inferno",
        "MAP05:_Minos%27_Judgement_(Master_Levels)",
    ),
    'nessus.wad': (
        7, 'Nessus: 5th Canto of Inferno', "MAP07:_Nessus_(Master_Levels)",
    ),
    'paradox.wad': (
        1, 'Paradox', 'MAP01:_Paradox_(Master_Levels)',
    ),
    'subspace.wad': (
        1, 'Subspace', 'MAP01:_Subspace_(Master_Levels)',
    ),
    'subterra.wad': (
        1, 'Subterra', 'MAP01:_Subterra_(Master_Levels)',
    ),
    'teeth.wad': (
        31,
        'The Express Elevator to Hell',
        'MAP31:_The_Express_Elevator_to_Hell_-_teeth.wad_(Master_Levels)',
    ),
    'teeth.wad*': (
        32,
        'Bad Dream',
        'MAP32:_Bad_Dream_-_teeth.wad_(Master_Levels)',
    ),
    'ttrap.wad': (
        1, 'Trapped on Titan', 'MAP01:_Trapped_on_Titan_(Master_Levels)',
    ),
    'vesperas.wad': (
        9, 'Vesperas: 7th Canto of Inferno', 'MAP09:_Vesperas_(Master_Levels)',
    ),
    'virgil.wad': (
        3,
        "Virgil's Lead: 3rd Canto of Inferno",
        "MAP03:_Virgil%27s_Lead_(Master_Levels)",
    ),
}

alternatives: list[str] = []
if os.path.islink('/etc/alternatives/doom'):
    # on Debian
    alternatives = [os.readlink('/etc/alternatives/doom')]

    proc = subprocess.check_output(
        ['update-alternatives', '--list', 'doom'],
        text=True,
    )
    for alternative in proc.splitlines():
        if alternative not in alternatives:
            alternatives.append(alternative)
else:
    # not on Debian
    for alternative in ('prboom-plus', 'prboom', 'chocolate-doom',
                        'crispy-doom', 'woof', 'dsda-doom'):
        if which(alternative):
            alternatives.append(alternative)


class MasterLevel(GObject.GObject):
    def __init__(
        self,
        game: str,
        warp: int,
        longname: str,
        url: str,
        description: str,
    ):
        super().__init__()
        self.game = game
        self.warp = warp
        self.longname = longname
        self.url = url
        self.description = description


class Launcher(Gtk.Application):
    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self.connect('activate', self.on_activate)

        self.game: str = ''
        self.warp: int = 0
        self.difficulty: int = 3
        self.engine: list[str] = []
        self.win: Gtk.ApplicationWindow
        self.games: Gio.ListStore

    def on_activate(self, app: 'Launcher') -> None:
        self.win = Gtk.ApplicationWindow(application=self)
        self.win.set_title("Doom 2 Master Levels")
        self.win.set_default_size(1020, 650)
        self.win.present()

        def warning_callback(self: Gtk.MessageDialog, data: int) -> None:
            exit(message)

        if len(alternatives) == 0:
            message = 'No DOOM engine found!\n' + requirements
            md = Gtk.MessageDialog(title='Warning',
                                   message_type=Gtk.MessageType.WARNING,
                                   buttons=Gtk.ButtonsType.CLOSE,
                                   text=message)
            md.set_transient_for(self.win)
            md.set_modal(True)
            md.connect("response", warning_callback)
            md.show()

        self.games = Gio.ListStore.new(MasterLevel)
        for k in sorted(levels.keys()):
            warp = levels[k][0]
            longname = levels[k][1]
            url = levels[k][2]
            level = os.path.splitext(k)[0]
            fullpath = DIR + '/%s.wad' % level
            if not os.path.isfile(fullpath):
                print('\n')
                message = fullpath + " is missing !\n" + requirements
                md = Gtk.MessageDialog(title='Warning',
                                       message_type=Gtk.MessageType.WARNING,
                                       buttons=Gtk.ButtonsType.CLOSE,
                                       text=message)
                md.set_transient_for(self.win)
                md.set_modal(True)
                md.connect("response", warning_callback)
                md.show()
            txt = '/usr/share/doc/doom2-masterlevels-wad/%s.txt' % level
            try:
                with open(txt, 'r', encoding='latin1') as f:
                    description = f.read()
            except (PermissionError, FileNotFoundError):
                description = "failed to read " + txt

            self.games.append(
                MasterLevel(level, warp, longname, url, description),
            )

        self.setup_ui()

    def setup_ui(self) -> None:
        grid = Gtk.Grid()
        grid.set_row_spacing(5)
        grid.set_column_spacing(5)
        self.win.set_child(grid)

        css_provider = Gtk.CssProvider()
        if Gtk.get_minor_version() < 12:
            css_provider.load_from_data(
                "label.smaller_row { margin-top: -5pt; margin-bottom: -2pt; }",
                -1,
            )
        else:
            css_provider.load_from_string(
                "label.smaller_row { margin-top: -5pt; margin-bottom: -2pt; }",
            )
        default_display = Gdk.Display.get_default()
        assert default_display is not None
        Gtk.StyleContext.add_provider_for_display(
            default_display, css_provider,
            Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION,
        )

        # level list
        def f_bind_setup(
            fact: Gtk.SignalListItemFactory,
            item: Gtk.ListItem,
        ) -> None:
            label = Gtk.Inscription()
            label.add_css_class("smaller_row")
            item.set_child(label)

        def f_bind_col1(
            fact: Gtk.SignalListItemFactory,
            item: Gtk.ListItem,
        ) -> None:
            level = item.get_item()
            assert isinstance(level, MasterLevel)
            label = item.get_child()
            assert isinstance(label, Gtk.Inscription)
            label.set_text(level.game)
            label.set_tooltip_text(level.longname)

        def f_bind_col2(
            fact: Gtk.SignalListItemFactory,
            item: Gtk.ListItem,
        ) -> None:
            level = item.get_item()
            assert isinstance(level, MasterLevel)
            label = item.get_child()
            assert isinstance(label, Gtk.Inscription)
            label.set_text(str(level.warp))
            label.set_tooltip_text(level.longname)

        selection = Gtk.SingleSelection.new(self.games)
        selection.connect("selection-changed", self.select_game)
        columnview = Gtk.ColumnView.new(selection)
        col1factory = Gtk.SignalListItemFactory()
        col2factory = Gtk.SignalListItemFactory()
        col1factory.connect("setup", f_bind_setup)
        col2factory.connect("setup", f_bind_setup)
        col1factory.connect("bind", f_bind_col1)
        col2factory.connect("bind", f_bind_col2)
        col1 = Gtk.ColumnViewColumn.new("Wad", col1factory)
        col1.set_fixed_width(80)
        col2 = Gtk.ColumnViewColumn.new("Map", col2factory)
        columnview.append_column(col1)
        columnview.append_column(col2)
        grid.attach(columnview, 0, 0, 1, 8)

        # header
        label = Gtk.Label()
        label.set_markup("<span size='xx-large'>Doom II Master Levels</span>")
        grid.attach(label, 1, 0, 1, 1)

        logo = Gtk.Image()
        logo.set_from_file(CACODEMON)
        logo.set_pixel_size(72)
        grid.attach(logo, 2, 0, 1, 1)

        # description
        scrolledwindow = Gtk.ScrolledWindow()
        scrolledwindow.set_policy(
            Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC,
        )
        grid.attach(scrolledwindow, 1, 1, 2, 1)

        self.textbuffer = Gtk.TextBuffer()
        self.textbuffer.set_text(
            'Please select a map from the list on the left',
        )

        textview = Gtk.TextView(buffer=self.textbuffer)
        textview.set_vexpand(True)
        textview.set_hexpand(True)
        textview.set_property('editable', False)
        textview.set_monospace(True)
        scrolledwindow.set_child(textview)

        self.doomwiki = Gtk.LinkButton(
            uri="https://doomwiki.org/wiki/Master_Levels_for_Doom_II",
            label="https://doomwiki.org/wiki/Master_Levels_for_Doom_II",
        )
        grid.attach(self.doomwiki, 1, 2, 2, 1)

        # difficulty
        difflabel = Gtk.Label(label="Choose your difficulty")
        grid.attach(difflabel, 1, 3, 1, 1)

        diffgrid = Gtk.Grid()
        diffradio = Gtk.CheckButton(label="1)  I'm too young to die")
        diffgrid.attach(diffradio, 0, 0, 1, 1)
        diffradio.connect('toggled', self.select_difficulty)
        for diff in ["2)  Hey, Not too Rough",
                     "3)  Hurt me plenty",
                     "4)  Ultra Violence",
                     "5)  Nightmare!"]:
            radiobutton = Gtk.CheckButton(label=diff)
            radiobutton.set_group(diffradio)
            radiobutton.connect('toggled', self.select_difficulty)
            if diff[0] == '3':
                radiobutton.set_active(True)
            diffgrid.attach(radiobutton, 0, int(diff[0]), 1, 1)
        grid.attach(diffgrid, 1, 4, 1, 1)

        # engine
        label = Gtk.Label(label="Choose your engine")
        grid.attach(label, 2, 3, 1, 1)
        radiogrid = Gtk.Grid()
        radiobuttonDefault = Gtk.CheckButton(label="n/a")
        radiobuttonDefault.set_visible(False)

        i = 0
        for alternative in alternatives:
            if alternative == '/usr/games/doomsday-compat':
                alternative = '/usr/games/doomsday'

            radiobutton = Gtk.CheckButton(label=alternative)
            radiobutton.set_group(radiobuttonDefault)
            radiobutton.connect('toggled', self.select_engine)
            radiogrid.attach(radiobutton, 0, i, 1, 1)

            if i == 0:
                radiobutton.set_label("%s (default)" % alternative)
                radiobutton.set_active(True)
                self.select_engine(radiobutton)

            # TODO: factorize these 3 stanzas & the 3 methods
            if (
                os.path.basename(alternative) == 'chocolate-doom'
                and which('chocolate-doom-setup')
            ):
                self.button_conf = Gtk.Button(label="Configure")
                radiogrid.attach(self.button_conf, 1, i, 1, 1)
                self.button_conf.connect("clicked", self.chocolate_setup)

            if (
                os.path.basename(alternative) == 'crispy-doom'
                and which('crispy-setup')
            ):
                self.button_conf = Gtk.Button(label="Configure")
                radiogrid.attach(self.button_conf, 1, i, 1, 1)
                self.button_conf.connect("clicked", self.crispy_setup)

            if (
                os.path.basename(alternative) == 'woof'
                and which('woof-setup')
            ):
                self.button_conf = Gtk.Button(label="Configure")
                radiogrid.attach(self.button_conf, 1, i, 1, 1)
                self.button_conf.connect("clicked", self.woof_setup)

            i += 1

        if i > 1 and os.path.isfile('/etc/debian_version'):
            radiogrid.set_tooltip_text(
                'Default can be changed with '
                '"update-alternatives --config doom"'
            )

        grid.attach(radiogrid, 2, 4, 1, 1)

        # Run !
        self.button_exec = Gtk.Button(label="Run")
        self.button_exec.set_sensitive(False)
        grid.attach(self.button_exec, 1, 6, 1, 1)
        self.button_exec.connect("clicked", self.run_game)

        button_quit = Gtk.Button(label="Exit")
        grid.attach(button_quit, 2, 6, 1, 1)
        button_quit.connect("clicked", lambda _: self.win.close())

    def select_game(
        self,
        selection: Gtk.SingleSelection,
        position: int,
        n_items: int,
    ) -> None:
        selected_item = selection.get_selected_item()
        assert isinstance(selected_item, MasterLevel)
        if selected_item is not None:
            self.button_exec.set_sensitive(True)
            self.game = selected_item.game
            self.warp = selected_item.warp
            self.textbuffer.set_text(selected_item.description)
            wad = self.game + '.wad'
            if self.game == 'teeth' and self.warp == 32:
                wad += '*'
            url = 'https://doomwiki.org/wiki/' + levels[wad][2]
            self.doomwiki.set_uri(url)
            self.doomwiki.set_label(url)

    def select_difficulty(self, radio: Gtk.CheckButton) -> None:
        self.difficulty = int(str(radio.get_label())[0])

    def select_engine(self, radio: Gtk.CheckButton) -> None:
        self.engine = [str(radio.get_label()).split(' ')[0]]
        if self.engine == ['/usr/games/doomsday']:
            self.engine.append('-game')
            self.engine.append('doom2')
        if self.engine == ['chocolate-doom'] and DIR == '/usr/share/doom':
            self.engine.append('-iwad')
            self.engine.append('/usr/share/doom/doom2.wad')

    def run_game(self, button: Gtk.Button) -> None:
        subprocess.call(
            self.engine + [
                '-file', '%s/%s.wad' % (DIR, self.game),
                '-warp', '%d' % self.warp,
                '-skill', '%d' % self.difficulty,
            ],
        )

    def chocolate_setup(self, button: Gtk.Button) -> None:
        subprocess.call(['chocolate-doom-setup'])

    def crispy_setup(self, button: Gtk.Button) -> None:
        subprocess.call(['crispy-setup'])

    def woof_setup(self, button: Gtk.Button) -> None:
        subprocess.call(['woof-setup'])


if __name__ == "__main__":
    launcher = Launcher(
        application_id="net.debian.game_data_packager.doom2_masterlevels",
    )
    launcher.run(None)



#!/usr/bin/python3
# encoding=utf-8
#
# Copyright © 2014-2015 Simon McVittie <smcv@debian.org>
# Copyright © 2015 Alexandre Detiste <alexandre@detiste.be>
# SPDX-License-Identifier: GPL-2.0-or-later

from __future__ import annotations

import errno
import os
import shlex
import shutil
import subprocess
import tarfile
import time
import zipfile
from abc import (ABCMeta, abstractmethod)
from collections.abc import (Collection, Iterator)
from types import TracebackType
from typing import (Any, BinaryIO)


class UnpackableEntry(metaclass=ABCMeta):
    """An entry in a StreamUnpackable.
    """
    @property
    @abstractmethod
    def is_directory(self) -> bool:
        raise NotImplementedError

    @property
    @abstractmethod
    def is_regular_file(self) -> bool:
        """True if the entry is a regular file. False if it is a
        directory, symlink, or some special thing like an instruction
        to patch some other file.
        """
        raise NotImplementedError

    @property
    def is_extractable(self) -> bool:
        """True if the entry is something that we can extract.

        The default implementation is that we can extract regular files.
        """
        return self.is_regular_file

    def get_symbolic_link_target(self) -> str | None:
        """Target of the symbolic link, or None if this is not a
        symbolic link.
        """
        return None

    @property
    def mtime(self) -> float | None:
        """The last-modification time, or None if unspecified."""
        return None

    @property
    @abstractmethod
    def name(self) -> str:
        """The absolute or relative filename, with Unix path separators."""
        raise NotImplementedError

    @property
    @abstractmethod
    def size(self) -> int:
        """The size in bytes."""
        raise NotImplementedError

    @property
    def type_indicator(self) -> str:
        """One or more ASCII symbols indicating the file type."""
        if self.is_directory:
            ret = 'd'
        elif self.is_regular_file:
            ret = '-'
        elif self.get_symbolic_link_target() is not None:
            ret = 'l'
        else:
            ret = '?'

        if self.is_extractable:
            ret += 'r'
        else:
            ret += '-'

        return ret


class SimpleUnpackable(metaclass=ABCMeta):
    """An archive in which entries can be inspected and extracted,
    but only all at once.
    """

    @property
    @abstractmethod
    def format(self) -> str:
        """Return the YAML "format" entry for this file.
        """
        raise NotImplementedError

    @abstractmethod
    def extractall(
        self,
        path: str,
        members: Collection[str] | None = None,
    ) -> None:
        """Extract all or most members of this archive to path.
        members is merely a hint: extracting more members than desired
        is allowed.
        """
        raise NotImplementedError

    def seekable(self) -> bool:
        """Return True if the unpacker is able to seek.
        """
        return False


class StreamUnpackable(SimpleUnpackable, metaclass=ABCMeta):
    """An archive in which entries can be inspected and extracted
    by iteration.
    """

    @abstractmethod
    def __iter__(self) -> Iterator[UnpackableEntry]:
        """Iterate through UnpackableEntry objects."""
        raise NotImplementedError

    @abstractmethod
    def open(self, member: str | UnpackableEntry) -> BinaryIO:
        """Open a binary file-like entry for the name or entry.
        """
        raise NotImplementedError

    def extract(
        self,
        member: str | UnpackableEntry,
        path: str | None = None,
    ) -> None:
        """Extract the given member from the archive into the given
        directory.
        """

        assert not isinstance(member, bytes), member

        if isinstance(member, str):
            filename = member
        else:
            filename = member.name

        reader = self.open(member)

        if not reader:
            raise ValueError('cannot open %s' % member)

        with reader:
            filename = filename.lstrip('/')

            while filename.startswith('../'):
                filename = filename[3:]
            filename = filename.replace('/../', '/')
            if filename.endswith('/..'):
                filename = filename[:-3]
            if filename.endswith('/'):
                filename = filename[:-1]
            if path is None:
                path = '.'

            dest = os.path.join(path, filename)
            os.makedirs(os.path.dirname(dest), exist_ok=True)

            try:
                os.remove(dest)
            except OSError as e:
                if e.errno != errno.ENOENT:
                    raise

            with open(dest, 'xb') as writer:
                shutil.copyfileobj(reader, writer)

    def extractall(
        self,
        path: str | None,
        members: Collection[str] | None = None,
    ) -> None:
        for entry in self:
            if entry.is_extractable:
                if members is None or entry.name in members:
                    self.extract(entry, path)

    def printdir(self) -> None:
        for entry in self:
            if entry.size is None:
                size = '?' * 9
            else:
                size = '%9s' % entry.size

            if entry.mtime is not None:
                mtime = time.strftime(
                    '%Y-%m-%d %H:%M:%S', time.gmtime(entry.mtime))
            else:
                mtime = '????-??-?? ??:??:??'

            print('%s %s %s %s' % (
                entry.type_indicator, size, mtime,
                shlex.quote(entry.name)))

    def seekable(self) -> bool:
        """Return True if the unpacker is able to seek.
        """
        return False

    def __enter__(self) -> StreamUnpackable:
        return self

    def __exit__(
        self,
        ex_type: type[BaseException] | None = None,
        ex_value: BaseException | None = None,
        ex_traceback: TracebackType | None = None
    ) -> None:
        pass


class WrapperUnpacker(StreamUnpackable):
    """Base class for a StreamUnpackable that wraps a TarFile-like object."""

    def __init__(self) -> None:
        # TODO: It must support open(), close() and iteration, but we don't
        # currently check that
        # + https://github.com/python/mypy/issues/3138
        self._impl: Any | None = None

    @abstractmethod
    def _wrap_entry(self, entry: Any) -> UnpackableEntry:
        raise NotImplementedError

    @abstractmethod
    def _is_entry(self, entry: Any) -> bool:
        raise NotImplementedError

    def __enter__(self) -> WrapperUnpacker:
        return self

    def __exit__(
        self,
        ex_type: type[BaseException] | None = None,
        ex_value: BaseException | None = None,
        ex_traceback: TracebackType | None = None
    ) -> None:
        if self._impl is not None:
            self._impl.close()
            self._impl = None

    def __iter__(self) -> Iterator[UnpackableEntry]:
        assert self._impl is not None, 'unpacker context not entered'

        for entry in self._impl:
            yield self._wrap_entry(entry)

    def open(self, entry: Any) -> BinaryIO:
        assert self._impl is not None, 'unpacker context not entered'
        assert self._is_entry(entry)
        contents = self._impl.open(entry.impl)
        assert contents is not None
        return contents


class TarEntry(UnpackableEntry):
    __slots__ = 'impl'

    def __init__(self, impl: tarfile.TarInfo) -> None:
        self.impl = impl

    @property
    def is_directory(self) -> bool:
        return self.impl.isdir()

    @property
    def is_regular_file(self) -> bool:
        return self.impl.isfile()

    @property
    def mtime(self) -> float:
        return self.impl.mtime

    @property
    def name(self) -> str:
        return self.impl.name

    @property
    def size(self) -> int:
        return self.impl.size

    def get_symbolic_link_target(self) -> str | None:
        if self.impl.issym():
            return self.impl.linkname
        else:
            return None

    @property
    def type_indicator(self) -> str:
        """One or more ASCII symbols indicating the file type."""
        if self.impl.isdir():
            ret = 'd'
        elif self.impl.isfile():
            ret = '-'
        elif self.impl.issym():
            ret = 'l'
        else:
            ret = '?<%r>' % self.impl.type

        if self.is_extractable:
            ret += 'r'
        else:
            ret += '-'

        return ret


class DpkgDebUnpacker(WrapperUnpacker):
    def __init__(self, path: str) -> None:
        self._path = path
        self._fsys_process: Any | None = None  # ?

    def __enter__(self) -> DpkgDebUnpacker:
        self._fsys_process = subprocess.Popen(
            ['dpkg-deb', '--fsys-tarfile', self._path],
            stdout=subprocess.PIPE,
        ).__enter__()
        assert self._fsys_process is not None
        self._impl = tarfile.open(
            self._path, mode='r|', fileobj=self._fsys_process.stdout,
        ).__enter__()
        return self

    def __exit__(
        self,
        ex_type: type[BaseException] | None = None,
        ex_value: BaseException | None = None,
        ex_traceback: TracebackType | None = None
    ) -> None:
        if self._impl is not None:
            self._impl.__exit__(ex_type, ex_value, ex_traceback)
            self._impl = None

        if self._fsys_process is not None:
            self._fsys_process.__exit__(ex_type, ex_value, ex_traceback)
            self._fsys_process = None

    @property
    def format(self) -> str:
        return 'deb'

    def open(self, entry: Any) -> BinaryIO:
        assert isinstance(entry, TarEntry)
        assert type(self._impl) is tarfile.TarFile
        contents = self._impl.extractfile(entry.impl)
        assert contents is not None
        return contents

    def _is_entry(self, entry: Any) -> bool:
        return isinstance(entry, TarEntry)

    def _wrap_entry(self, entry: tarfile.TarInfo) -> UnpackableEntry:
        return TarEntry(entry)


class TarUnpacker(WrapperUnpacker):
    def __init__(
        self,
        name: str,
        reader: BinaryIO | None = None,
        compression: str = '*',
        skip: int = 0,
    ) -> None:
        super(TarUnpacker, self).__init__()
        self.skip = skip
        self.compression = compression

        if reader is None:
            reader = open(name, 'rb')

        if skip:
            discard = reader.read(skip)
            assert len(discard) == skip

        self._impl = tarfile.open(
            name, mode='r|' + compression, fileobj=reader)

    @property
    def format(self) -> str:
        return 'tar.' + self.compression

    def open(self, entry: Any) -> BinaryIO:
        assert isinstance(entry, TarEntry)
        assert type(self._impl) is tarfile.TarFile
        contents = self._impl.extractfile(entry.impl)
        assert contents is not None
        return contents

    def _is_entry(self, entry: Any) -> bool:
        return isinstance(entry, TarEntry)

    def _wrap_entry(self, entry: tarfile.TarInfo) -> UnpackableEntry:
        return TarEntry(entry)


class ZipEntry(UnpackableEntry):
    __slots__ = 'impl'

    def __init__(self, impl: zipfile.ZipInfo) -> None:
        self.impl = impl

    @property
    def is_directory(self) -> bool:
        return self.name.endswith('/')

    @property
    def is_regular_file(self) -> bool:
        return not self.name.endswith('/')

    @property
    def mtime(self) -> float:
        return time.mktime(self.impl.date_time + (0, 0, -1))

    @property
    def name(self) -> str:
        return self.impl.filename

    @property
    def size(self) -> int:
        return self.impl.file_size


class ZipUnpacker(WrapperUnpacker):
    def __init__(
        self,
        file_or_name: str | BinaryIO
    ) -> None:
        super(ZipUnpacker, self).__init__()
        if hasattr(file_or_name, 'seekable') and not file_or_name.seekable():
            self.__seekable = False
        else:
            # zip files based on an on-disk file are seekable
            self.__seekable = True

        self._impl = zipfile.ZipFile(file_or_name, 'r')

    def __iter__(self) -> Iterator[ZipEntry]:
        assert type(self._impl) is zipfile.ZipFile
        for entry in self._impl.infolist():
            yield ZipEntry(entry)

    def _is_entry(self, entry: Any) -> bool:
        return isinstance(entry, ZipEntry)

    def _wrap_entry(self, entry: zipfile.ZipInfo) -> UnpackableEntry:
        return ZipEntry(entry)

    @property
    def format(self) -> str:
        return 'zip'

    def seekable(self) -> bool:
        return self.__seekable



#!/usr/bin/python3
# encoding=utf-8
#
# Copyright © 2015 Alexandre Detiste <alexandre@detiste.be>
# SPDX-License-Identifier: GPL-2.0-or-later

# pre-calculate a list of packages from yaml files
# for bash auto-completion

import os
import glob

srcdir = os.path.dirname(
    os.path.dirname(
        os.path.abspath(__file__)
    )
)

for yaml_file in sorted(glob.glob(os.path.join(srcdir, 'data', '*.yaml'))):
    print(os.path.splitext(os.path.basename(yaml_file))[0])
    with open(yaml_file, encoding='utf-8') as raw:
        for line in raw:
            if line.strip() == 'packages:':
                break
        for line in raw:
            if line == '\n':
                continue
            if line[0] != ' ':
                break
            if line[2] not in (' ', '#'):
                print(line.strip(':\n '), end=' ')

        print('')



#!/usr/bin/python3
# encoding=utf-8
#
# Copyright © 2024 Alexandre Detiste <alexandre@detiste.be>
# SPDX-License-Identifier: GPL-2.0-or-later

import os
import subprocess

from game_data_packager.game import load_games
from game_data_packager.games.scummvm_common import (ScummvmGameData,
                                                     ScummvmPackage)


def get_scummvm_ids() -> set[str]:
    result = set()
    stdout = subprocess.check_output(['scummvm', '--list-all-games'],
                                     text=True)
    for line in stdout.splitlines():
        if ':' in line:
            gameid = line.split()[0]
            result.add(gameid)
    return result


scummvm_ids = get_scummvm_ids()

scummvm_ids2: dict[str, int] = dict()
for gameid in scummvm_ids:
    assert gameid not in scummvm_ids2
    scummvm_ids2[gameid] = 1

    oldgameid = gameid.split(':')[1]
    scummvm_ids2[oldgameid] = scummvm_ids2.get(oldgameid, 0) + 1

scummvm_ids3: set[str] = set()
for gameid, count in scummvm_ids2.items():
    if count > 1:
        # print('discarding non-discriminating id: %s' % gameid)
        pass
    else:
        scummvm_ids3.add(gameid)


def get_gdp_ids() -> set[str]:
    games = set()
    for name, game in load_games(
        datadir=os.environ.get('GDP_BUILDDIR', 'out')
    ).items():
        if game.wikibase != 'https://wiki.scummvm.org/index.php/':
            continue
        assert isinstance(game, ScummvmGameData), game
        game.load_file_data()
        for package in game.packages.values():
            assert isinstance(package, ScummvmPackage), package
            gameid = package.gameid or game.gameid
            games.add(gameid)
    return games


gdp_ids = get_gdp_ids()
# print(gdp_ids)

print("Unknown game id's:")
print(sorted(gdp_ids - scummvm_ids3))



#!/usr/bin/python3
# encoding=utf-8
#
# Copyright © 2016 Alexandre Detiste <alexandre@detiste.be>
# SPDX-License-Identifier: GPL-2.0-or-later

# this is just a proof-of-concept,
# please don't start packaging 10.000 DOS games

from __future__ import annotations

import configparser
import logging
import os
from typing import (Any, TYPE_CHECKING)

from ..build import (PackagingTask)
from ..data import (Package)
from ..game import (GameData)
from ..util import (mkdir_p)

if TYPE_CHECKING:
    from typing import (Unpack)
    from ..packaging import (PerPackageState, PackagingTaskArgs)

logger = logging.getLogger(__name__)


class DosboxGameData(GameData):
    """Special subclass of GameData for DOS games.

    These games will need the "dosgame" runtime
    provided by src:game-data-packager.
    """

    def __init__(self, shortname: str, data: dict[str, Any]) -> None:
        super(DosboxGameData, self).__init__(shortname, data)
        self.binary_executables = 'all'

    def construct_package(
        self,
        binary: str,
        data: dict[str, Any]
    ) -> DosboxPackage:
        return DosboxPackage(binary, data)

    def construct_task(
        self,
        **kwargs: Unpack[PackagingTaskArgs]
    ) -> DosboxTask:
        return DosboxTask(self, **kwargs)


class DosboxPackage(Package):
    def __init__(self, binary: str, data: dict[str, Any]) -> None:
        super(DosboxPackage, self).__init__(binary, data)

        assert 'install_to' not in data
        assert 'depends' not in data

        self.install_to = '$assets/dosbox/' + self.name[:len(self.name)-5]
        self.depends = 'dosgame'
        self.main_exe: str
        if 'main_exe' in data:
            self.main_exe = data['main_exe']
            return

        for wanted in self.install:
            filename, ext = os.path.splitext(wanted)
            if (
                filename not in ('config', 'install', 'setup')
                and ext in ('.com', '.exe')
            ):
                self.main_exe = filename
                return


class DosboxTask(PackagingTask):
    def fill_extra_files(
        self,
        per_package_state: PerPackageState
    ) -> None:
        super().fill_extra_files(per_package_state)
        package = per_package_state.package
        destdir = per_package_state.destdir
        assert isinstance(package, DosboxPackage)

        pgm = package.name[:len(package.name)-5]
        bindir = self.packaging.substitute(self.packaging.BINDIR, package.name)
        mkdir_p(os.path.join(destdir, bindir.strip('/')))
        os.symlink('dosgame', os.path.join(destdir, bindir.strip('/'), pgm))

        appdir = os.path.join(destdir, 'usr/share/applications')
        mkdir_p(appdir)

        desktop = configparser.RawConfigParser()
        desktop.optionxform = lambda option: option     # type: ignore
        desktop['Desktop Entry'] = {}
        entry = desktop['Desktop Entry']
        entry['Name'] = package.longname or self.game.longname
        entry['Icon'] = 'dosbox'
        if self.game.genre is None:
            entry['GenericName'] = 'game'
        else:
            entry['GenericName'] = self.game.genre + ' game'
        entry['Exec'] = pgm
        entry['Terminal'] = 'false'
        entry['Type'] = 'Application'
        entry['Categories'] = 'Game;'

        with open(
            os.path.join(appdir, '%s.desktop' % package.name),
            'w', encoding='utf-8',
        ) as output:
            desktop.write(output, space_around_delimiters=False)

        # minimal information needed by the runtime
        dosgame = configparser.RawConfigParser()
        dosgame.optionxform = lambda option: option     # type: ignore
        dosgame['Dos Game'] = {}
        entry = dosgame['Dos Game']
        entry['Dir'] = package.main_exe
        entry['Exe'] = package.main_exe

        install_to = self.packaging.substitute(
            package.install_to, package.name,
        )
        with open(
            os.path.join(destdir, install_to.strip('/'), 'dosgame.inf'),
            'w', encoding='utf-8',
        ) as output:
            dosgame.write(output, space_around_delimiters=False)


GAME_DATA_SUBCLASS = DosboxGameData



#!/usr/bin/python3
# encoding=utf-8
#
# Copyright © 2015 Simon McVittie <smcv@debian.org>
# SPDX-License-Identifier: GPL-2.0-or-later

import textwrap

if __name__ == '__main__':
    cin = open('debian/copyright.in', encoding='utf-8')
    copy = open('debian/copyright', 'w', encoding='utf-8')

    copy.write(cin.__next__())
    copy.write('# This file was generated by debian/copyright.py ')
    copy.write('from debian/copyright.in\n')
    copy.write('# Please edit one of those files instead\n')

    for line in cin:
        if line.lower().startswith('include-license:'):
            license = line.split(':', 1)[1].strip()
            with open(license, encoding='utf-8') as lic:
                for line in lic:
                    if line.strip():
                        copy.write('  ' + line)
                    else:
                        copy.write(' .\n')
        elif line.lower().startswith('include-license-wrap:'):
            license = line.split(':', 1)[1].strip()
            with open(license, encoding='utf-8') as lic:
                for line in lic:
                    if line.strip():
                        for outline in textwrap.wrap(
                            line,
                            break_long_words=False,
                            initial_indent=' ',
                            subsequent_indent=' ',
                            tabsize=2,
                        ):
                            copy.write(outline + '\n')
                    else:
                        copy.write(' .\n')
        else:
            copy.write(line)
    cin.close()
    copy.close()



#!/usr/bin/python3
# encoding=utf-8
#
# Copyright © 2015-2016 Simon McVittie <smcv@debian.org>
# Copyright © 2015-2016 Alexandre Detiste <alexandre@detiste.be>
# SPDX-License-Identifier: GPL-2.0-or-later

from __future__ import annotations

import configparser
import logging
import os
import subprocess
import urllib.error
import urllib.request
from collections.abc import (Iterable, Iterator)
from typing import (Any, TYPE_CHECKING)

from ..build import (PackagingTask)
from ..data import (Package, PackageRelation)
from ..game import (GameData)
from ..paths import DATADIR
from ..util import (mkdir_p)
from ..version import (GAME_PACKAGE_VERSION)

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from typing import Unpack
    from ..packaging import (PerPackageState, PackagingTaskArgs)


def install_data(from_: str, to: str) -> None:
    subprocess.check_call(['cp', '--reflink=auto', from_, to])


class ScummvmPackage(Package):
    gameid: str | None


class ScummvmGameData(GameData):
    def __init__(self, shortname: str, data: dict[str, Any]) -> None:
        super(ScummvmGameData, self).__init__(shortname, data)

        self.wikibase = 'https://wiki.scummvm.org/index.php/'
        assert self.wiki

        if 'gameid' in self.data:
            self.gameid: str = self.data['gameid']
            assert self.gameid != shortname, \
                'extraneous gameid for ' + shortname
            self.aliases.add(self.gameid)
        else:
            self.gameid = shortname
        assert self.gameid

        if self.engine is None:
            self.engine = 'scummvm'
        if self.genre is None:
            self.genre = 'Adventure'

    def construct_package(self, binary: str, data: dict[str, Any]) -> Package:
        return ScummvmPackage(binary, data)

    def _populate_package(
        self,
        package: Package,
        d: dict[str, Any]
    ) -> None:
        super(ScummvmGameData, self)._populate_package(package, d)
        assert isinstance(package, ScummvmPackage), package
        package.gameid = d.get('gameid')

    def construct_task(
        self,
        **kwargs: Unpack[PackagingTaskArgs]
    ) -> ScummvmTask:
        return ScummvmTask(self, **kwargs)


class ScummvmTask(PackagingTask):
    def iter_extra_paths(
        self,
        packages: Iterable[Package]
    ) -> Iterator[str]:
        super(ScummvmTask, self).iter_extra_paths(packages)

        gameids = set()
        for p in packages:
            assert isinstance(self.game, ScummvmGameData), self.game
            assert isinstance(p, ScummvmPackage), p
            gameid = p.gameid or self.game.gameid
            if gameid == 'agi-fanmade':
                continue
            gameids.add(gameid.split('-')[0])
        if not gameids:
            return

        # http://wiki.scummvm.org/index.php/User_Manual/Configuring_ScummVM
        # https://github.com/scummvm/scummvm/pull/656
        config_home = os.environ.get(
            'XDG_CONFIG_HOME',
            os.path.expanduser('~/.config'),
        )
        for rcfile in (
            os.path.join(config_home, 'scummvm/scummvm.ini'),
            os.path.expanduser('~/.scummvmrc'),
            os.path.join(config_home, 'residualvm/residualvm.ini'),
            os.path.expanduser('~/.residualvmrc'),
        ):
            if os.path.isfile(rcfile):
                config = configparser.ConfigParser(strict=False)
                config.read(rcfile, encoding='utf-8')
                for section in config.sections():
                    for gameid in gameids:
                        if section.split('-')[0] == gameid:
                            if 'path' not in config[section]:
                                # invalid .scummvmrc
                                continue
                            path = config[section]['path']
                            if os.path.isdir(path):
                                yield path

    def download_icon(
        self,
        per_package_state: PerPackageState
    ) -> str | None:
        '''this is "Fair Use", the user already owns the game assets'''
        scummvm_toc = 'https://raw.githubusercontent.com/' + \
                      'scummvm/scummvm-icons/master/TOC.md'

        package = per_package_state.package
        assert isinstance(package, ScummvmPackage), package
        game = self.game
        assert isinstance(game, ScummvmGameData), game
        dest_icon = os.path.join(self.get_workdir(), package.name + '.png')
        gameid = package.gameid or game.gameid
        dlfilename = None

        if not per_package_state.download:
            return None

        if ':' in gameid:
            parts = gameid.split(':')
            dlfilename = '%s-%s.png' % (parts[0], parts[1])
        else:
            try:
                response = urllib.request.urlopen(scummvm_toc)
            except urllib.error.HTTPError:
                logger.warning("Can't download %s", scummvm_toc)
                return None

            for line in response.read().decode('utf-8').split("\n"):
                if f"| {gameid} " in line:
                    parts = (line.split('|'))
                    parts = [s.strip() for s in parts]
                    if parts[1] == "✅":
                        dlfilename = '%s-%s.png' % (parts[2], parts[3])
                    else:
                        dlfilename = '%s.png' % parts[2]
                    break

        if dlfilename is None:
            logger.debug("Can't find an icon to download for '%s'", gameid)
            return None

        url = 'https://raw.githubusercontent.com/' + \
              'scummvm/scummvm-icons/master/icons/%s' % dlfilename
        try:
            logger.debug("Downloading icon %s", url)
            response = urllib.request.urlopen(url)
        except urllib.error.HTTPError:
            logger.warning("Can't download %s", url)
            return None

        with open(dest_icon, 'wb') as w:
            w.write(response.read())

        return dest_icon

    def fill_extra_files(
        self,
        per_package_state: PerPackageState
    ) -> None:
        super().fill_extra_files(per_package_state)
        package = per_package_state.package
        destdir = per_package_state.destdir
        assert isinstance(self.game, ScummvmGameData), self.game
        assert isinstance(package, ScummvmPackage), package

        if package.type not in ('demo', 'full') and not package.gameid:
            return
        if package.data_type == 'binaries':
            return
        if package.name.endswith('-music'):
            return
        if package.name.endswith('-video'):
            return

        assert package.name.endswith('-data'), package.name

        package.relations['recommends'].append(
            PackageRelation(
                'game-data-packager-runtime (>= %s~)' % GAME_PACKAGE_VERSION
            )
        )

        icon = package.name[0:len(package.name)-len('-data')]
        pixdir = os.path.join(destdir, 'usr/share/pixmaps')
        for from_ in (self.locate_steam_icon(package),
                      os.path.join(pixdir, self.get_extra_icon_name(package)),
                      os.path.join(DATADIR, package.name + '.png'),
                      os.path.join(DATADIR, self.game.shortname + '.png'),
                      os.path.join('/usr/share/pixmaps', icon + '.png'),
                      os.path.join(
                          DATADIR,
                          self.game.shortname.strip('1234567890') + '.png'),
                      self.download_icon(per_package_state),
                      ):
            if from_ and os.path.exists(from_):
                mkdir_p(pixdir)
                install_data(from_, os.path.join(pixdir, '%s.png' % icon))
                break
        else:
            svgdir = '/usr/share/icons/hicolor/scalable/apps'
            symlinkdir = '$assets/game-data-packager-runtime'
            symlinksrc = os.path.join(symlinkdir, 'scummvm.svg')
            symlinkdest = os.path.join(svgdir, '%s.svg' % icon)
            package.symlinks[symlinkdest] = symlinksrc
            from_ = symlinkdest

        assert from_ is not None

        from_ = os.path.splitext(from_)[0] + '.svgz'
        if os.path.exists(from_):
            svgdir = os.path.join(
                destdir, 'usr/share/icons/hicolor/scalable/apps',
            )
            mkdir_p(svgdir)
            install_data(from_, os.path.join(svgdir, '%s.svgz' % icon))

        appdir = 'usr/share/applications'

        symlinkdir = '$assets/game-data-packager-runtime/scummvm'
        symlinksrc = os.path.join(symlinkdir, '%s.desktop' % icon)
        symlinkdest = os.path.join(appdir, '%s.desktop' % icon)
        package.symlinks[symlinkdest] = symlinksrc


GAME_DATA_SUBCLASS = ScummvmGameData



#!/usr/bin/python3
# encoding=utf-8
#
# Copyright © 2016 Alexandre Detiste <alexandre@detiste.be>
# SPDX-License-Identifier: GPL-2.0-or-later

# this goes in "/usr/games/dosgame"

import configparser
import os
import sys
import subprocess

game = os.path.basename(sys.argv[0])

srcroot = '/usr/share/games/dosbox/'
srcdir = os.path.join(srcroot, game)
inf = os.path.join(srcdir, 'dosgame.inf')

if game == 'dosgame':
    print('Supported games:')
    print('\n'.join(os.listdir(srcroot)))
    exit(0)

config = configparser.ConfigParser()
config.read(inf, encoding='utf-8')
for section in config.sections():
    # this .inf file could also include
    # other sections that would be
    # copied as-is into dosbox.cfg
    assert section == "Dos Game"
    dir = config[section]['Dir']
    exe = config[section]['Exe']

destroot = os.path.expanduser('~/.dosbox')
destdir = os.path.join(destroot, dir)
autoexec = os.path.join(destdir, 'dosbox.cfg')

# XXX: currenlty only work with games that
#      have all assets in a single directory
#
#      some games needs to be able to write in subdirs
#      and will need extensive linkfarms,
#      while other are ok with symlinked subdirs
if not os.path.isdir(destdir):
    os.makedirs(destdir)
    for dirpath, dirnames, filenames in os.walk(srcdir):
        for fn in filenames:
            if fn == 'dosgame.inf':
                continue
            full = os.path.join(dirpath, fn)
            os.symlink(full, os.path.join(destdir, fn))

if not os.path.isfile(autoexec):
    with open(autoexec, 'w', encoding='ascii') as of:
        of.write('[Autoexec]\n')
        of.write('@ECHO OFF\n')
        of.write('MOUNT C %s\n' % destroot)
        of.write('C:\n')
        of.write('CD %s\n' % dir)
        of.write('%s\n' % exe)
        of.write('EXIT\n')

subprocess.call(['dosbox', '-conf', autoexec])



# Copyright 2023 Alexandre Detiste
# SPDX-License-Identifier: FSFAP

# this goes in /usr/lib/python3/dist-packages/omg/__init__.pyi
# + touch /usr/lib/python3/dist-packages/omg/py.typed

from PIL.Image import Image

class Graphic:

    def to_Image(self) -> Image: ...

class WAD:

    graphics: dict[str, Graphic]

    def __init__(self, from_file:str): ...



#!/usr/bin/python3
# encoding=utf-8
# Copyright 2024 Simon McVittie
# SPDX-License-Identifier: FSFAP

import os
import sys
import unittest
from pathlib import Path
from typing import Any


TOP_SRCDIR = Path(__file__).resolve().parent.parent


if 'GDP_UNINSTALLED' in os.environ:
    sys.path.insert(0, str(TOP_SRCDIR))
else:
    sys.path.insert(0, '/usr/share/game-data-packager')
    sys.path.insert(0, '/usr/share/games/game-data-packager')


class AssertVersionMatches(unittest.TestCase):
    def setUp(self) -> None:
        pass

    def test_version_number(self) -> None:
        from game_data_packager.version import (
            GAME_PACKAGE_VERSION,
            GAME_PACKAGE_RELEASE,
        )

        contents: dict[str, Any] = {}

        if 'GDP_BUILDDIR' in os.environ:
            TOP_BUILDDIR = Path(os.environ['GDP_BUILDDIR']).resolve()
            with open(TOP_BUILDDIR / 'version.py') as reader:
                exec(reader.read(), contents, contents)

        try:
            from debian.changelog import Changelog
        except ImportError as e:
            self.skipTest(str(e))

        with open(TOP_SRCDIR / 'debian' / 'changelog') as reader:
            changelog = Changelog(reader, max_blocks=1)

        print(f'# version in debian/changelog: {changelog.version!r}')
        print(f'# dists in debian/changelog: {changelog.distributions!r}')
        print(f'# version from build system: {GAME_PACKAGE_VERSION!r}')
        print(f'# revision from build system: {GAME_PACKAGE_RELEASE!r}')

        if 'UNRELEASED' in changelog.distributions:
            self.skipTest('package not yet released')

        if '~' in str(changelog.version) or '+' in str(changelog.version):
            self.skipTest(
                "version number has ~ or +, assuming you know what "
                "you're doing"
            )

        if 'ubuntu' in str(changelog.version):
            self.skipTest("this test is skipped on Ubuntu and derivatives")

        self.assertEqual(
            changelog.upstream_version,
            GAME_PACKAGE_VERSION,
        )
        self.assertEqual(
            changelog.debian_revision or '',
            GAME_PACKAGE_RELEASE,
        )

        if 'GDP_BUILDDIR' in os.environ:
            GAME_PACKAGE_VERSION = contents['GAME_PACKAGE_VERSION']
            GAME_PACKAGE_RELEASE = contents['GAME_PACKAGE_RELEASE']

            print(f'# version from built files: {GAME_PACKAGE_VERSION!r}')
            print(f'# revision from built files: {GAME_PACKAGE_RELEASE!r}')

            self.assertEqual(
                changelog.upstream_version,
                GAME_PACKAGE_VERSION,
            )
            self.assertEqual(
                changelog.debian_revision or '',
                GAME_PACKAGE_RELEASE,
            )

    def tearDown(self) -> None:
        pass


if __name__ == '__main__':
    from gdp_test_common import main
    main()



#!/usr/bin/python3
# encoding=utf-8
#
# Copyright © 2015 Simon McVittie <smcv@debian.org>
# SPDX-License-Identifier: GPL-2.0-or-later

from __future__ import annotations

import logging
import os
from typing import TYPE_CHECKING, Any

from ..build import (PackagingTask)
from ..game import GameData
from ..util import TemporaryUmask, mkdir_p

if TYPE_CHECKING:
    import argparse
    from typing import Unpack
    from ..packaging import (PerPackageState, PackagingTaskArgs)

logger = logging.getLogger(__name__)


class QuakeTask(PackagingTask):
    """With hindsight, it would have been better to make the TryExec
    in quake point to a symlink to the quake executable, or something;
    but with a name like hipnotic-tryexec.sh it would seem silly for it
    not to be a shell script.
    """

    def fill_extra_files(
        self,
        per_package_state: PerPackageState
    ) -> None:
        super().fill_extra_files(per_package_state)
        package = per_package_state.package
        destdir = per_package_state.destdir

        for path in package.install:
            if path.startswith('hipnotic'):
                detector = 'hipnotic-tryexec.sh'
                break
            elif path.startswith('rogue'):
                detector = 'rogue-tryexec.sh'
                break
        else:
            return

        with TemporaryUmask(0o022):
            quakedir = os.path.join(destdir, 'usr/share/games/quake')
            mkdir_p(quakedir)
            path = os.path.join(quakedir, detector)
            with open(path, 'w') as f:
                f.write('#!/bin/sh\nexit 0\n')
            os.chmod(path, 0o755)


class QuakeGameData(GameData):
    def construct_task(self, **kwargs: Unpack[PackagingTaskArgs]) -> QuakeTask:
        return QuakeTask(self, **kwargs)

    def add_parser(
        self,
        parsers: argparse._SubParsersAction[Any],
        base_parser: argparse.ArgumentParser,
        **kwargs: Any,
    ) -> argparse.ArgumentParser:
        parser = super(QuakeGameData, self).add_parser(
            parsers, base_parser, conflict_handler='resolve',
        )
        parser.add_argument(
            '-m', dest='packages',
            action='append_const', const='quake-registered',
            help='Equivalent to --package=quake-registered',
        )
        parser.add_argument(
            '-s', dest='packages',
            action='append_const', const='quake-shareware',
            help='Equivalent to --package=quake-shareware',
        )
        parser.add_argument(
            '--mp1', '-mp1', dest='packages',
            action='append_const', const='quake-armagon',
            help='Equivalent to --package=quake-armagon',
        )
        parser.add_argument(
            '--mp2', '-mp2', dest='packages',
            action='append_const', const='quake-dissolution',
            help='Equivalent to --package=quake-dissolution',
        )
        parser.add_argument(
            '--music', dest='packages',
            action='append_const', const='quake-music',
            help='Equivalent to --package=quake-music',
        )
        parser.add_argument(
            '--mp1-music', dest='packages',
            action='append_const', const='quake-armagon-music',
            help='Equivalent to --package=quake-armagon-music',
        )
        parser.add_argument(
            '--mp2-music', dest='packages',
            action='append_const', const='quake-dissolution-music',
            help='Equivalent to --package=quake-dissolution-music',
        )
        return parser


GAME_DATA_SUBCLASS = QuakeGameData



#!/usr/bin/python3
# encoding=utf-8
#
# Copyright © 2015 Simon McVittie <smcv@debian.org>
# Copyright © 2015 Alexandre Detiste <alexandre@detiste.be>
# SPDX-License-Identifier: GPL-2.0-or-later

from __future__ import annotations

import logging
import os
import subprocess
import sys
import tarfile
import tempfile
import glob
import zipfile
from collections.abc import (Iterable, Iterator)
from shutil import which
from typing import (Any, BinaryIO, TYPE_CHECKING)

import yaml
try:
    from debian.deb822 import Deb822
    ON_DEBIAN = True
except ImportError:
    ON_DEBIAN = False

from .data import (HashedFile, PackageRelation, WantedFile)
from .game import (GameData, load_game, load_games)
from .gog import GOG
from .steam import parse_acf
from .unpack import (DpkgDebUnpacker, TarUnpacker)
from .unpack.auto import automatic_unpacker
from .unpack.innoextract import (InnoSetup)
from .util import (
        check_output,
        rm_rf,
        )

if TYPE_CHECKING:
    import argparse
    import io
    from .data import (FileGroup, Package)
    from .unpack import (StreamUnpackable)

logging.basicConfig()
logger = logging.getLogger(__name__)


def guess_lang(string: str) -> str | None:
    string = string.lower()
    path = os.path.basename(string.rstrip('/'))
    if path.split('-')[-1] in ('de', 'en', 'es', 'fr', 'it', 'ja', 'pl', 'ru'):
        return path
    for short, long in [('de', 'german'),
                        ('es', 'spanish'),
                        ('fr', 'french'),
                        ('it', 'italian'),
                        ('ja', 'japanese'),
                        ('pl', 'polish'),
                        ('ru', 'russian')]:
        if long in string:
            return short
    # check against other args
    for lang in (
        'german',
        'spanish',
        'french',
        'italian',
        'polish',
        'russian',
    ):
        for arg in sys.argv:
            if lang in arg:
                return 'en'
    return None


def is_probably_optional(file: str) -> bool:
    file = file.split('?')[0]
    name, ext = os.path.splitext(file.lower())
    if ext in ('cfg', 'cmd', 'com', 'drv', 'ico', 'ini'):
        return True
    return False


def is_license(file: str) -> bool:
    file = file.split('?')[0]
    name, ext = os.path.splitext(file.lower())
    if ext not in ('.doc', '.htm', '.html', '.pdf', '.txt', ''):
        return False
    for word in ('eula', 'license', 'vendor'):
        if word in name:
            return True
    return False


def iter_path_components(filename: str) -> Iterator[str]:
    dirname, basename = os.path.split(filename)

    if dirname == filename:
        # the root
        yield dirname
    elif dirname:
        for x in iter_path_components(dirname):
            yield x

    if basename:
        yield basename


def is_doc(file: str) -> bool:
    file = file.split('?')[0]

    for dirname in iter_path_components(file):
        if dirname.lower() in ('manual', 'docs', 'doc', 'help'):
            return True

    name, ext = os.path.splitext(file.lower())

    if ext not in ('.doc', '.htm', '.html', '.pdf', '.txt', ''):
        return False

    for word in ('changes', 'hintbook', 'manual', 'quickstart',
                 'readme', 'refcard', 'reference', 'support'):
        if word in name:
            return True

    return False


def is_dosbox(
    file: str,
    opened: BinaryIO | None = None
) -> bool:
    '''check if DOSBox assests are just dropped in games assets directory'''
    basename = os.path.basename(file)
    if basename in ('dosbox.conf',  'dosbox.exe',
                    'dosbox-0.71.tar.gz', 'dosbox-0.74.tar.gz',
                    'SDL_net.dll', 'SDL.dll', 'zmbv.dll', 'zmbv.inf'):
        return True
    # to check: COPYING.txt INSTALL.txt NEWS.txt THANKS.txt *.conf
    if basename.startswith('dosbox'):
        return True
    if basename not in ('AUTHORS.txt', 'README.txt'):
        return False

    txt: BinaryIO | io.BufferedReader
    if opened is None:
        txt = open(file, 'rb')
    else:
        txt = opened

    # We can't use a TextIOWrapper because tarfile's _Stream objects
    # don't implement seekable(), which breaks _FileInFile's
    # implementation of seekable().
    line = txt.readline()
    return 'dosbox' in line.decode('latin1').lower()


def is_runtime(path: str) -> bool:
    dir_l = path.lower()
    for runtime in ('data.now', 'directx', 'dosbox'):
        if '/%s/' % runtime in dir_l:
            return True
        if dir_l.endswith('/' + runtime):
            logger.warning('ignoring %s runtime at %s' % (runtime, path))
            return True
    return False


class Template:
    def __init__(
        self,
        game: GameData,
        strip_paths: Iterable[str] = ()
    ) -> None:
        self.game = game
        self.plugin: str | None = None
        self.strip_paths = strip_paths
        self.has_dosbox: bool = False
        self.preexisting_files = set(self.game.files.keys())

    def new_group(self, stem: str) -> FileGroup:
        i = 0
        name = stem

        while name in self.game.files or name in self.game.groups:
            i += 1
            name = '%s (%d)' % (stem, i)

        return self.game.ensure_group(name)

    def is_scummvm(self, path: str) -> bool:
        dir_l = path.lower()
        if dir_l.endswith('/scummvm') or '/scummvm/' in dir_l:
            self.plugin = 'scummvm_common'
            return True
        return False

    def add_archive(
        self,
        path: str,
        lower: bool = False,
        reader: Any = None,  # not used
        unpack: bool = True,
    ) -> WantedFile:
        out_name = os.path.basename(path)

        if lower:
            out_name = out_name.lower()

        result = self.add_file(path, out_name=out_name, unpack=unpack)
        self.game.ensure_group('archives').group_members.add(result.name)
        return result

    def add_unpacker(
        self,
        path: str,
        result: WantedFile,
        unpacker: StreamUnpackable
    ) -> None:
        if result.unpack is None:
            result.unpack = dict(format=unpacker.format)

        if isinstance(unpacker, TarUnpacker):
            result.unpack['skip'] = unpacker.skip

        contents = []

        for entry in unpacker:
            if entry.is_regular_file and entry.is_extractable:
                try:
                    provided = self.add_file(
                        entry.name,
                        opened=unpacker.open(entry),
                        size=entry.size,
                        parent_unpacker=unpacker)
                except NotImplementedError as e:
                    logger.warning("Can't decompress %s from %s: %s" %
                                   (entry.name, result.name, e))
                else:
                    contents.append(provided)

        self.reconcile_groups(
            contents, stem=result.name, provider=result,
        )

    def reconcile_groups(
        self,
        files: Iterable[HashedFile],
        stem: str | None = None,
        provider: WantedFile | None = None,
        package: Package | None = None,
    ) -> None:
        names = set()
        remaining = set()
        groups_found = {}
        optional = set()

        for f in files:
            names.add(f.name)
            remaining.add(f.name)

        # If there are pre-existing groups that are strict subsets of
        # what's left to classify, then take them out, largest first.
        if remaining:
            weighted = []

            for name, g in self.game.groups.items():
                weighted.append((len(g.group_members), g.name, g))

            weighted.sort(reverse=True)

            for _, _, g in weighted:
                if g.group_members <= remaining:
                    remaining -= g.group_members
                    groups_found[g.name] = g

        # If there are pre-existing groups with a non-trivial
        # intersection with what's left to classify, then take those,
        # largest intersection first.
        while remaining:
            weighted = []

            for name, g in self.game.groups.items():
                intersection = g.group_members & remaining

                # Arbitrarily decide that overlaps of less than 5 are
                # uninteresting
                if len(intersection) >= 5:
                    weighted.append((len(intersection), g.name, g))

            if not weighted:
                break

            weighted.sort(reverse=True)

            for _, _, g in weighted:
                intersecting_group = self.new_group(
                    '%s in %s' % (g.name, stem),
                )
                intersecting_group.group_members = g.group_members & names
                groups_found[intersecting_group.name] = intersecting_group
                disjoint_group = self.new_group(
                    '%s not in %s' % (g.name, stem),
                )
                disjoint_group.group_members = g.group_members - names

                remaining -= intersecting_group.group_members

        # If there's anything left, create new groups.
        if remaining:
            prefix = 'remaining ' if groups_found else ''

            main_group = self.new_group('%scontents of %s' % (prefix, stem))
            opt_group = self.new_group(
                '%scontents of %s - optional' % (prefix, stem),
            )
            optional.add(opt_group.name)
            doc_group = self.new_group(
                '%scontents of %s - documentation' % (prefix, stem),
            )
            doc_group.doc = True
            license_group = self.new_group(
                '%scontents of %s - licenses' % (prefix, stem),
            )
            license_group.license = True
            # only from .deb
            abs_group = self.new_group(
                '%scontents of %s - absolute paths' % (prefix, stem),
            )
            abs_group.install_to = '.'
            optional.add(abs_group.name)

            for n in remaining:
                f = self.game.files[n]

                if f.install_to == '.':
                    abs_group.group_members.add(n)
                elif f.license and f.doc:
                    license_group.group_members.add(n)
                    doc_group.group_members.add(n)
                elif f.license:
                    license_group.group_members.add(n)
                elif f.doc:
                    doc_group.group_members.add(n)
                elif f.ignorable or is_probably_optional(f.name):
                    opt_group.group_members.add(n)
                else:
                    main_group.group_members.add(n)

            for g in (
                main_group, opt_group, doc_group, license_group, abs_group,
            ):
                if g.group_members:
                    groups_found[g.name] = g

        # Check that the set of names hasn't changed during this process
        union = set()

        for g in groups_found.values():
            for m in g.group_members:
                union.add(m)

        assert union == names

        if provider:
            provider.provides = set()

            for name, g in groups_found.items():
                if g.group_members:
                    provider.provides.add(name)

            provider.provides_files = set(
                self.game.iter_expand_groups(provider.provides))

        if package:
            for name, g in groups_found.items():
                if g.group_members:
                    if g.license or g.doc or g.ignorable or name in optional:
                        package.optional.add(name)
                    else:
                        package.install.add(name)

    def add_file(
        self,
        path: str,
        out_name: str | None = None,
        opened: BinaryIO | None = None,
        size: int | None = None,
        lang: str | None = None,
        parent_unpacker: StreamUnpackable | None = None,
        unpack: bool = False,
    ) -> WantedFile:
        ignorable = False

        if out_name is None:
            out_name = path

        if is_dosbox(path, opened):
            self.has_dosbox = True
            ignorable = True
        elif os.path.splitext(path.lower())[1] in (
            '.exe', '.exe$0', '.ovl',
            '.dll', '.dll$0', '.bat', '.386',
        ):
            ignorable = True
        elif out_name.startswith('goggame-') or out_name in (
            'webcache.zip', 'gog.ico', 'gfw_high.ico',
        ):
            ignorable = True

        if opened is None:
            is_plain_file = True
            opened = open(path, 'rb')
        else:
            is_plain_file = False

        hf = HashedFile.from_file(path, opened, size=size)
        return self.add_hashed_file(
            path, hf,
            out_name=out_name, lang=lang,
            parent_unpacker=parent_unpacker, size=size,
            is_plain_file=is_plain_file, unpack=unpack, opened=opened,
            ignorable=ignorable,
        )

    def add_hashed_file(
        self,
        path: str,
        hf: HashedFile,
        *,
        out_name: str | None = None,
        lang: str | None = None,
        parent_unpacker: StreamUnpackable | None = None,
        size: int | None = None,
        unpack: bool = False,
        is_plain_file: bool = False,
        opened: BinaryIO | None = None,
        ignorable: bool = False,
    ) -> WantedFile:
        if out_name is None:
            out_name = path

        result = None
        existing = None
        match_path = '/' + path.lower()

        # TODO: This resembles PackagingTask.consider_file()
        # and PackagingTask.use_file()
        for look_for, candidates in self.game.known_filenames.items():
            if match_path.endswith('/' + look_for):
                for c in candidates:
                    existing = self.game.files[c]

                    if hf.matches(existing):
                        # TODO: Show provenance of archive members somehow
                        logger.info('Found %s at %s', existing.name, out_name)

                        if existing.size is None:
                            existing.size = hf.size

                        if existing.md5 is None:
                            existing.md5 = hf.md5

                        if existing.sha1 is None:
                            existing.sha1 = hf.sha1

                        if existing.sha256 is None:
                            existing.sha256 = hf.sha256

                        result = existing
                        break
                    else:
                        existing = None

                if result is not None:
                    break
        else:
            # We are creating a new WantedFile
            assert existing is None
            assert hf.md5 is not None

            for prefix in self.strip_paths:
                if out_name.startswith(prefix + '/'):
                    out_name = out_name[len(prefix) + 1:]

            if out_name not in self.game.files:
                pass
            elif lang and out_name + '?' + lang not in self.game.files:
                out_name += ('?' + lang)

            elif out_name + '?' + hf.md5[:6] not in self.game.files:
                out_name += ('?' + hf.md5[:6])

            else:
                i = 0

                while out_name + '?' + str(i) in self.game.files:
                    i += 1

                out_name += ('?' + str(i))

            assert out_name not in self.game.files

            result = WantedFile(out_name)
            result.size = hf.size
            result.md5 = hf.md5
            result.sha1 = hf.sha1
            result.sha256 = hf.sha256

            if ignorable:
                result.ignorable = True

            self.game.files[result.name] = result

        assert type(result) is WantedFile

        for lf in result.look_for:
            self.game.known_filenames.setdefault(lf, set()).add(result.name)

        if result.md5 is not None:
            self.game.known_md5s.setdefault(result.md5, set()).add(result.name)

        if result.sha1 is not None:
            self.game.known_sha1s.setdefault(result.sha1, set()).add(
                result.name)

        if result.sha256 is not None:
            self.game.known_sha256s.setdefault(result.sha256, set()).add(
                result.name)

        unpacker = None

        if unpack and (parent_unpacker is None or parent_unpacker.seekable()):
            # skip PK3s: we don't normally want to recurse into them
            if not path.endswith('.pk3'):
                if is_plain_file:
                    unpacker = automatic_unpacker(path)
                else:
                    unpacker = automatic_unpacker(path, opened)

        if result is not existing and ignorable:
            result.ignorable = True

        if result is not existing and is_license(path):
            result.license = True

        if result is not existing and is_doc(path):
            result.doc = True

        if unpacker is not None:
            with unpacker:
                self.add_unpacker(path, result, unpacker)

        return result

    def add_template(
        self,
        template: str,
        name: str,
        lower: bool = False
    ) -> None:
        template_game = load_game(
            False, template, None, name=name, yaml_file=template)
        template_game.load_file_data()
        files = {}

        package_name = 'template-%s' % os.path.basename(template)
        i = 0

        while package_name in self.game.packages:
            i += 1
            package_name = 'template-%s-unique%d' % (
                os.path.basename(template), i)

        package = self.game.construct_package(package_name, {})
        self.game.packages[package_name] = package

        for name, f in template_game.files.items():
            files[name] = self.add_hashed_file(
                f.filename, f,
                out_name=name, is_plain_file=False, unpack=False,
                size=f.size,
            )

        for g in template_game.groups.values():
            members = []

            for name in g.group_members:
                if name in files:
                    members.append(files[name])

            self.reconcile_groups(
                members,
                stem='group "' + g.name + '" in ' + os.path.basename(template),
            )

        ungrouped = []

        for name, f in template_game.files.items():
            if f.provides:
                members = []

                for member in f.provides_files:
                    members.append(files[member.name])

                self.reconcile_groups(
                    members,
                    stem='files provided by "%s" in %s' % (
                        name, os.path.basename(template)
                    ),
                    provider=files[f.name],
                )

            for g in template_game.groups.values():
                if name in g.group_members:
                    break
            else:
                ungrouped.append(files[name])

            if f.unpack:
                files[name].unpack = f.unpack

        self.reconcile_groups(
            ungrouped,
            stem='no particular group in ' + os.path.basename(template),
        )

        self.reconcile_groups(
            list(files.values()),
            stem='files listed in ' + os.path.basename(template),
            package=package,
        )

    def add_one_dir(
        self,
        destdir: str,
        lower: bool = False,
        game: str | None = None,
        lang: str | None = None,
        group_stem: str | None = None,
    ) -> None:
        basename = os.path.basename(os.path.abspath(destdir))

        if group_stem is None:
            group_stem = basename

        if destdir.startswith('/usr/local') or destdir.startswith('/opt/'):
            self.game.try_repack_from.append(destdir)

        if not game:
            game = basename

        if game.endswith('-data'):
            game = game[:len(game) - 5]

        steam = max(destdir.find('/SteamApps/common/'),
                    destdir.find('/steamapps/common/'))
        if steam > 0:
            steam_dict: dict[str, int | str] = dict()
            steam_id = 'FIXME'
            for acf in parse_acf(destdir[:steam+11]):
                if '/common/' + acf['installdir'] in destdir:
                    steam_id = acf['appid']
                    self.game.longname = game = acf['name']
                    break
            steam_dict['id'] = int(steam_id)
            steam_dict['path'] = destdir[steam+11:]

        virtual = None
        package = None
        assert game
        game = game.replace(' ', '').replace(':', '').replace('_', '-').lower()

        if lang:
            stem = package_name = game + '-' + lang + '-data'
            virtual = game + '-data'
        else:
            stem = package_name = game + '-data'

        i = 0

        while package_name in self.game.packages:
            i += 1
            package_name = '%s-unique%d' % (stem, i)

        package = self.game.construct_package(package_name, {})
        self.game.packages[package_name] = package

        if lang:
            if lang != 'en':
                package.langs = [lang]

        if virtual:
            package.relations['provides'].append(PackageRelation(virtual))

        if steam > 0:
            package.steam = steam_dict

        contents = []

        for dirpath, dirnames, filenames in os.walk(destdir):
            if self.is_scummvm(dirpath) or is_runtime(dirpath):
                continue

            for fn in filenames:
                path = os.path.join(dirpath, fn)

                assert path.startswith(destdir + '/')
                name = path[len(destdir) + 1:]
                out_name = name
                if lower:
                    out_name = out_name.lower()

                if os.path.isdir(path):
                    continue
                elif os.path.islink(path):
                    package.symlinks[path] = os.path.realpath(path)
                elif os.path.isfile(path):
                    contents.append(
                        self.add_file(path, out_name=out_name, lang=lang))
                else:
                    logger.warning('ignoring unknown file type at %s' % path)

            if self.has_dosbox:
                logger.warning(
                    'DOSBOX files detected, make sure not to include those '
                    'in your package'
                )

        if self.plugin != 'scummvm_common':
            package.install_to = '$assets/' + game

        self.reconcile_groups(contents, group_stem, package=package)

    def add_one_gog_sh(self, archive: str) -> None:
        self.add_archive(archive, unpack=True)

        with zipfile.ZipFile(archive, 'r') as zf:
            if 'scripts/config.lua' in zf.namelist():
                with zf.open('scripts/config.lua') as metadata:
                    for line in metadata.read().decode().splitlines():
                        line = line.strip()
                        if line.startswith('id = '):
                            self.game.gog['path'] = '"%s"' % line.split('"')[1]

    def add_one_innoextract(self, exe: str, lower: bool) -> None:
        game = self.game.gog['game'] = GOG.get_id_from_archive(exe)
        if not game:
            game = os.path.basename(exe)
            game = game[len('setup_'):len(game)-len('.exe')]
            last_part = game.split('_')[-1]
            if last_part.strip('0123456789.') == '':
                game = game[0:len(game)-len(last_part)-1]
            last_part = game.split('_')[-1]
            if last_part in (
                'german',
                'spanish',
                'french',
                'italian',
                'polish',
                'russian',
            ):
                game = game[0:len(game)-len(last_part)-1]

        tmp = tempfile.mkdtemp(prefix='gdptmp.')

        result = self.add_archive(exe, lower, unpack=False)
        logger.info('Unpacking "%s" with innoextract...', exe)

        with InnoSetup(os.path.realpath(exe)) as unpacker:
            log = unpacker._extractall_for_template(tmp)

        self.game.longname = log.split('\n')[0].split('"')[1]

        self.add_one_dir(
            tmp, lang=guess_lang(exe), lower=lower, group_stem=result.name)
        rm_rf(tmp)

        result.unpack = dict(format='innoextract')
        result.provides = set()

    def add_one_deb(self, deb: str, lower: bool) -> None:
        if not ON_DEBIAN or not which('dpkg-deb'):
            exit('.deb analysis is only implemented on Debian.')

        control = None

        result = self.add_archive(deb, unpack=False)

        version = None
        with subprocess.Popen(
            ['dpkg-deb', '--ctrl-tarfile', deb],
            stdout=subprocess.PIPE,
        ) as ctrl_process:
            with tarfile.open(
                deb + '//control.tar.*',
                mode='r|',
                fileobj=ctrl_process.stdout,
            ) as ctrl_tarfile:
                for ctrl_entry in ctrl_tarfile:
                    name = ctrl_entry.name
                    if name == '.':
                        continue

                    if name.startswith('./'):
                        name = name[2:]
                    if name == 'control':
                        reader = ctrl_tarfile.extractfile(ctrl_entry)
                        control = Deb822(reader)
                        print('# data/%s.control.in' % control['package'])
                        version = control['version']
                        if 'Homepage' in control:
                            if 'gog.com/' in control['Homepage']:
                                self.game.gog['url'] = (
                                    control['Homepage'].split('/')[-1]
                                )

                        control.dump(fd=sys.stdout, text_mode=True)
                        print('')
                    elif name == 'preinst':
                        logger.warning('ignoring preinst, not supported yet')
                    elif name == 'md5sums':
                        pass
                    else:
                        logger.warning('unknown control member: %s', name)

        if control is None:
            exit('Could not find DEBIAN/control')

        result.unpack = dict(format='deb')

        package_name = control['package']
        i = 0

        while package_name in self.game.packages:
            i += 1
            package_name = '%s-unique%d' % (control['package'], i)

        package = self.game.construct_package(package_name, {})
        self.game.packages[package_name] = package

        if version:
            package.version = version

        install_to = None
        contents = []

        with DpkgDebUnpacker(deb) as unpacker:
            for entry in unpacker:
                name = entry.name
                if name.startswith('./'):
                    name = name[2:]

                if self.is_scummvm(name) or is_runtime(name):
                    continue
                if (name.startswith('usr/share/doc/') and
                        name.endswith('changelog.gz')):
                    continue
                if (name.startswith('usr/share/doc/') and
                        name.endswith('changelog.Debian.gz')):
                    continue

                if (name.startswith('usr/share/doc/') and
                        name.endswith('copyright')):
                    print('# data/%s.copyright' % control['package'])
                    for line in unpacker.open(entry):
                        print(line.decode('utf-8'), end='')
                    print('')
                    continue

                if entry.is_regular_file and install_to is None:
                    # assume this is the place
                    if name.startswith('usr/share/games/'):
                        there = name[len('usr/share/games/'):]
                        there = there.split('/', 1)[0]
                        install_to = ('usr/share/games/' + there)
                    elif name.startswith('opt/GOG Games/'):
                        there = name[len('opt/GOG Games/'):]
                        there = there.split('/', 1)[0]
                        install_to = ('opt/GOG Games/' + there)
                        self.game.gog['path'] = '"%s"' % there

                target = entry.get_symbolic_link_target()

                if entry.is_regular_file:
                    abs_path = False
                    doc = False
                    ignorable = False
                    license = False

                    if name.startswith('usr/share/doc/'):
                        name = name[len('usr/share/doc/'):]
                        name = name.split('/', 1)[1]
                        doc = True

                    name_l = name.lower()
                    basename_l = os.path.basename(name_l)

                    if os.path.splitext(name_l)[1] in ('.exe', '.bat'):
                        ignorable = True
                    elif (
                        'support/gog' in name_l
                        or os.path.basename(name_l) in (
                            'start.sh', 'uninstall.sh'
                        )
                    ):
                        ignorable = True
                    elif name.startswith('opt/') and is_license(name):
                        name = basename_l
                        license = True
                    elif name.startswith('opt/') and is_doc(name):
                        doc = True
                    elif (not doc and
                            install_to is not None and
                            name.startswith(install_to + '/')):
                        name = name[len(install_to) + 1:]
                        if lower:
                            name = name.lower()
                        if self.game.gog and name.startswith('data/'):
                            name = name[len('data/'):]
                    elif not doc:
                        abs_path = True

                    provided = self.add_file(
                        deb + '//data.tar.*//' + name,
                        out_name=name,
                        opened=unpacker.open(entry),
                        size=entry.size,
                        parent_unpacker=unpacker)

                    if provided.name not in self.preexisting_files:
                        if doc:
                            provided.doc = True

                        if license:
                            provided.license = True

                        if ignorable:
                            provided.ignorable = True

                        if abs_path:
                            provided.install_to = '.'

                    contents.append(provided)
                elif entry.is_directory:
                    pass
                elif target is not None:
                    package.symlinks[name] = os.path.join(
                        os.path.dirname(name), target)
                else:
                    logger.warning(
                        'unhandled data.tar entry type: %s: %s',
                        name, entry.type_indicator,
                    )

        if self.plugin != 'scummvm_common' and install_to is not None:
            package.install_to = os.path.join(
                '/', install_to
            ).replace('/usr/share/games/', '$assets/')

        self.reconcile_groups(
            contents, stem=result.name,
            package=package, provider=result,
        )

    def print_yaml(self) -> None:
        print('---')

        if self.plugin:
            print('plugin: %s' % self.plugin)
            print('')

        data = self.game.to_data(expand=False, include_ignorable=True)

        if data:
            yaml.dump(data, default_flow_style=False, stream=sys.stdout)

        print('\n...')


def do_one_exec(pgm: list[str], lower: bool) -> None:
    print('running:', pgm)
    with subprocess.Popen(
        ['strace', '-e', 'open', '-s', '100'] + pgm,
        stderr=subprocess.PIPE,
        stdout=subprocess.DEVNULL,
        text=True,
    ) as proc:
        used = set()
        missing = set()
        while proc.poll() is None:
            stderr = proc.stderr
            assert stderr is not None
            line = stderr.readline().strip()
            if not line.startswith('open('):
                continue
            file = line.split('"')[1]
            file = file.replace('//', '/')
            if file.startswith('/usr/share/scummvm'):
                continue
            if not (
                file.startswith('/usr/share/games')
                or file.startswith('/usr/share/' + pgm[0])
                or file.startswith('/usr/local/')
            ):
                continue
            if 'ENOENT' in line:
                missing.add(file)
            else:
                used.add(file)

        dirs = set()
        print('# used')
        for file in sorted(used):
            dirs.add(os.path.dirname(file))
            print("    - %s" % file)
        if missing:
            print('# missing ?')
            for file in sorted(missing):
                print("    - %s" % file)

        present = set()
        for dir in dirs:
            for dirpath, dirnames, filenames in os.walk(dir):
                for fn in filenames:
                    present.add(os.path.join(dirpath, fn))

        unused = present - used
        if unused:
            print('# not used')
            for file in sorted(unused):
                print("    - %s" % file)


def do_flacsums(destdir: str, lower: bool) -> None:
    if not which('ffmpeg'):
        exit('Install ffmpeg')
    if not which('metaflac'):
        exit('Install metaflac')

    fla_or_flac = '.fla'
    md5s: dict[str, str] = dict()
    done_wav = 0
    done_flac = 0
    for filename in glob.glob(os.path.join(destdir, '*')):
        file = os.path.basename(filename).lower()
        file, ext = os.path.splitext(file)
        if ext == '.wav':
            md5 = check_output(
                ['ffmpeg', '-i', filename, '-f', 'md5', '-'],
                stderr=subprocess.DEVNULL,
                text=True,
            )
            md5 = md5.rstrip().split('=')[1]
            assert file not in md5s or md5s[file] == md5, \
                   "got different md5's for %s.wav|flac" % file
            md5s[file] = md5
            done_wav += 1
        if ext == '.flac':
            fla_or_flac = '.flac'
        if ext in ('.fla', '.flac'):
            md5 = check_output(
                ['metaflac', '--show-md5sum', filename],
                text=True,
            )
            md5 = md5.rstrip()
            assert file not in md5s or md5s[file] == md5, \
                   "got different md5's for %s.wav|flac" % file
            md5s[file] = md5
            done_flac += 1

    if not md5s:
        exit("Couldn' find any .wav or .flac file")

    print('flacsums: |')
    for file in sorted(md5s.keys()):
        print('  %s  %s' % (md5s[file], file + fla_or_flac))

    print("\n#processed %i .wav and %i .fla[c] files" % (done_wav, done_flac))


def main(
    args: argparse.Namespace,
    games_: dict[str, GameData]  # not used
) -> None:
    # ./run make-template -e -- scummvm -p /usr/share/games/spacequest1/ sq1
    if args.execute:
        do_one_exec(args.args, args.lower)
        return

    if args.flacsums:
        do_flacsums(args.args[0], args.lower)
        return

    if args.base is None:
        game = GameData(
            '__template__',
            dict(
                copyright='© 1970 FIXME',
                packages={},
            ),
        )
    else:
        games = load_games(args.base)

        if args.base not in games:
            raise SystemExit('Unable to load game %r' % args.base)

        game = games[args.base]

    game.load_file_data()
    template = Template(game, strip_paths=args.strip_paths)

    for t in args.templates:
        template.add_template(t, lower=args.lower, name=game.shortname)

    # "./run make-template setup_<game>.exe gog_<game>.deb"
    # will merge files lists
    for arg in args.args:
        basename = os.path.basename(arg)
        if os.path.isdir(arg):
            template.add_one_dir(
                arg.rstrip('/'), args.lower, lang=guess_lang(arg),
            )
        elif arg.endswith('.deb'):
            template.add_one_deb(arg, args.lower)
        elif basename.startswith('setup_') and arg.endswith('.exe'):
            if not which('innoextract'):
                exit('Install innoextract')
            template.add_one_innoextract(arg, lower=args.lower)
        elif basename.startswith('gog_') and arg.endswith('.sh'):
            template.add_one_gog_sh(arg)
        else:
            template.add_archive(arg, lower=args.lower, unpack=True)

    template.print_yaml()



#!/usr/bin/python3
# encoding=utf-8
#
# Copyright © 2015 Alexandre Detiste <alexandre@detiste.be>
# SPDX-License-Identifier: GPL-2.0-or-later

# Usage:
# out/run-tool-uninstalled tools/babel.py
# rsync -P out/available.html
#          illusion.pseudorandom.co.uk:/srv/game-data-packager.debian.net/www/available.html

# Online at https://game-data-packager.debian.net/available.html

from __future__ import annotations

import os

from game_data_packager.build import (FillResult)
from game_data_packager.game import (load_games)

SHOPS = [
   ('url_steam', 'Steam',
    'https://steamcommunity.com/groups/debian_gdp#curation'),
   ('url_gog', 'GOG.com', None),
   ('url_misc', 'Misc.', None)
]


class Game:
    todo: bool = False
    total: int = 0
    demos: int = 0
    langs: dict[str, int]
    missing_langs: set[str]
    multi_langs: set[str]
    genre: str = 'Unknown'
    year: int | None = None
    franchise: str | None = None
    shortname: str
    longname: str
    # free-as-in-beer
    fullfree: bool = True
    somefree: bool = False
    url_wiki: str | None = None
    url_wikipedia: str | None = None
    url_steam: str | None = None
    url_gog: str | None = None
    url_misc: str | None = None

    def __init__(self, shortname: str) -> None:
        self.shortname = ''
        for char in shortname.lower():
            if 'a' <= char <= 'z' or '0' <= char <= '9':
                self.shortname += char
        self.langs = dict()
        self.missing_langs = set()
        self.multi_langs = set()

    def __repr__(self) -> str:
        return '<%s:%s>' % (self.genre, self.shortname)

    def __lt__(self, other: Game) -> bool:
        if self.genre < other.genre:
            return True
        elif self.genre > other.genre:
            return False

        if self.franchise and self.year:
            self_sort_key = '%s %d%s' % (
                self.franchise.lower(), self.year, self.shortname,
            )
        else:
            self_sort_key = self.shortname

        if other.franchise and other.year:
            other_sort_key = '%s %d%s' % (
                other.franchise.lower(), other.year, other.shortname,
            )
        else:
            other_sort_key = other.shortname

        return self_sort_key < other_sort_key


def load_stats() -> list[Game]:
    games: list[Game] = []

    for name, game in load_games().items():
        g = Game(name)
        g.genre = game.genre or 'Unkown'
        g.franchise = game.franchise
        g.longname = game.longname
        g.missing_langs = set(game.missing_langs)
        g.url_wikipedia = game.wikipedia
        g.url_wiki = game.wikibase + (game.wiki or '')
        g.url_steam = game.url_steam
        g.url_gog = game.url_gog
        g.url_misc = game.url_misc

        if game.copyright:
            g.year = int(game.copyright[2:6])

        for package in game.packages.values():
            g.total += 1

            if package.copyright:
                year = int(package.copyright[2:6])
                if not g.year:
                    g.year = year
                else:
                    g.year = min(g.year, year)

            if package.demo_for:
                g.demos += 1
            else:
                g.langs[package.lang] = g.langs.get(package.lang, 0) + 1
                g.multi_langs.update(set(package.langs))

            with game.construct_task() as task:
                if task.fill_gaps(
                    package=package, log=False,
                ) is FillResult.IMPOSSIBLE:
                    if not package.better_versions:
                        g.fullfree = False
                else:
                    g.somefree = True

        games.append(g)

    return games


games = load_stats()


# add missing games from list
def load_todo() -> list[Game]:
    games: list[Game] = list()

    with open('debian/TODO', 'r', encoding='utf8') as missing:
        for line in missing:
            if line[0:2] == '##':
                break

        genre = 'Unknown'
        for line in missing:
            line = line.strip()
            if line[0:1] == '#':
                genre = line[1:len(line)].strip()
                continue
            elif line == '':
                continue

            if '#' in line:
                longname, bug = line.split('#')
                if ' ' in bug:
                    bug, remainder = bug.split(maxsplit=1)
                else:
                    remainder = ''
                shortname = longname
                longname = (
                    '%s <a href="https://bugs.debian.org/'
                    'cgi-bin/bugreport.cgi?bug=%s">#%s</a> %s'
                ) % (longname, bug, bug, remainder)
            elif 'http://' in line or 'https://' in line:
                shortname = line
                longname = line   # TODO
            else:
                shortname = line
                longname = line

            g = Game(shortname)
            g.todo = True
            g.fullfree = False
            g.genre = genre
            g.longname = longname

            games.append(g)
    return games


games += load_todo()

# compute totals
langs: dict[str, int] = dict()
genres: dict[str, int] = dict()
for game in games:
    genres[game.genre] = genres.get(game.genre, 0) + 1
    for lang, count in game.langs.items():
        langs[lang] = langs.get(lang, 0) + count

games.sort()

output = os.path.join(os.environ.get('GDP_BUILDDIR', 'out'), 'available.html')

html = open(output, 'w', encoding='utf8')
html.write('''\
<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN"
 "http://www.w3.org/TR/html4/loose.dtd">
<html>
<head>
<title>Game-Data-Packager</title>
<meta http-equiv="Content-Type" content="text/html;charset=utf-8">
<meta name="generator" \
content="https://salsa.debian.org/games-team/game-data-packager/\
tree/master/tools/babel.py">
<style type="text/css">
''')

all_langs = sorted(langs.keys())
for lang in all_langs:
    if lang == 'en':
        continue
    html.write('  #check-%s:checked ~ * .%s' % (lang, lang))
    if lang != all_langs[-1]:
        html.write(',\n')

html.write(''' {
  background-color: yellow;
}
</style>
</head>

<body>
<h1>Debian Games Team</h1>
<img src="../proposed-logo.png" height="64" width="64"
 alt="Debian Games Team logo">
<h2>List of games supported by <code>game-data-packager</code> in git</h2>
This is an automatically generated list of games supported by the
upcoming release.<br>
Please visit the
<a href="http://wiki.debian.org/Games/GameDataPackager">Wiki</a>
for more general information.
<br><br>
''')

langs_order = [
    k
    for k, v in sorted(langs.items(), key=lambda kv: (-kv[1], kv[0]))
]
for lang in langs_order:
    if lang == 'en':
        continue
    html.write(
        '<!--label for=check-%s-->%s<!--/label-->'
        '<input id=check-%s type=checkbox>&nbsp;&nbsp;'
        % (lang, lang, lang)
    )

html.write('''<table border=1 cellspacing=0>
<tr>
<td colspan=2>&nbsp</td>
<td>WP</td>
<td>yaml</td>
<td><b>Total</b></td>
''')

for lang in langs_order:
    html.write('  <td><b>%s</b></td>\n' % lang)
html.write('<td>Demo</td>')
for _, name, url in SHOPS:
    if url:
        html.write('<td><a href="%s">%s</a></td>' % (url, name))
    else:
        html.write('<td>%s</td>' % name)
html.write('</tr>')


# BODY
def body() -> tuple[int, int]:
    last_genre = None
    total = 0
    demos = 0
    wikipedia = set()
    for game in games:
        total += game.total
        html.write('<tr>\n')
        genre = game.genre
        if genre != last_genre:
            html.write('<td rowspan=%i>%s</td>\n' % (genres[genre], genre))
            last_genre = genre
        highlight = [
            lang
            for lang in langs_order
            if lang != 'en' and lang in game.langs
        ]
        highlight += list(game.missing_langs)
        if highlight:
            css = ' class="%s"' % ' '.join(highlight)
        else:
            css = ''
        html.write('  <td%s>' % css)
        if game.url_wiki:
            html.write('<a href="%s">%s</a>' % (game.url_wiki, game.longname))
        else:
            html.write(game.longname)
        html.write('</td>')

        wp = game.url_wikipedia
        if wp in wikipedia:
            html.write('<td><b>&Prime;</b></td>')
        elif wp:
            html.write('<td><a href="%s"><b>W</b></a></td>' % wp)
            wikipedia.add(wp)
        else:
            html.write('<td>&nbsp;</td>')

        if game.todo:
            html.write('<td>&nbsp;</td>')
            html.write('  <td bgcolor="orange">!</td>\n')
        else:
            html.write('<td><a href="https://salsa.debian.org/games-team/'
                       'game-data-packager/tree/master/data/%s.yaml">'
                       '⌨</a></td>' % game.shortname)
            html.write('  <td bgcolor="lightgreen">%s</td>\n' % game.total)

        for lang in langs_order:
            count = game.langs.get(lang, 0)
            if count and lang == 'en':
                html.write('  <td bgcolor="lightgreen">%s</td>\n' % count)
            elif count:
                html.write('  <td bgcolor="green">%s</td>\n' % count)
            elif lang in game.multi_langs:
                html.write('  <td bgcolor="lightgreen">*</td>\n')
            elif lang in game.missing_langs:
                html.write('  <td bgcolor="orange">!</td>\n')
            else:
                html.write('  <td>&nbsp;</td>\n')

        if game.fullfree:
            html.write(
                '  <td colspan=%d align=center%s><b>freeload</b></td>\n'
                % (len(SHOPS) + 1, css)
            )
        else:
            if game.demos:
                demos += game.demos
                html.write('  <td align=center><b>%i</b></td>\n' % game.demos)
            elif game.somefree:
                html.write('  <td align=center><b>X</b></td>\n')
            else:
                html.write('  <td>&nbsp;</td>\n')
            for u in (game.url_steam, game.url_gog, game.url_misc):
                if u:
                    html.write(
                        '  <td align=center%s><a href="%s"><b>X</b></a></td>\n'
                        % (css, u)
                    )
                else:
                    html.write('  <td%s>&nbsp;</td>\n' % css)
        html.write('</tr>\n')

    return (total, demos)


total, demos = body()

# TOTAL
html.write('<tr><td colspan=4><b>Total</b></td>\n')
html.write('  <td><b>%i</b></td>\n' % total)
for lang in langs_order:
    html.write('  <td><b>%i</b></td>\n' % langs[lang])

html.write('  <td><b>%i</b></td>\n' % demos)

html.write('''<td colspan=%s>&nbsp;</td>
</tr>
</table>
<ul>
<li>! : language is missing</li>
<li>* : multi-lang support in a single package</li>
</ul>
</body>
</html>
''' % len(SHOPS)
)

html.close()



#!/usr/bin/env python3
# encoding=utf-8
#
# Copyright © 2024 Sébastien Noel <sebastien@twolife.be>
# SPDX-License-Identifier: GPL-2.0-or-later
import os
import sys
import sdl2
import sdl2.ext
from enum import Enum
from PIL import Image, ImageFile
from sdl2 import sdlmixer
from sdl2.ext import FontTTF

if os.path.isdir('/usr/share/smac'):
    SMACDIR = '/usr/share/smac'
elif os.path.isdir('/usr/share/games/smac'):
    SMACDIR = '/usr/share/games/smac'
else:
    SMACDIR = os.path.dirname(os.path.realpath(__file__))


def data_filename(filename: str) -> str:
    return os.path.join(
                SMACDIR,
                'data',
                filename,
            )


def is_in_rect(rect: sdl2.rect.SDL_Rect, x: int, y: int) -> bool:
    return (x >= rect.x and (x < rect.x + rect.w)) and \
           (y >= rect.y and (y < rect.y + rect.h))


class Button(Enum):
    SMAC = 1
    SMACX = 2
    CANCEL = 3


class FlcFile(Enum):
    FLARE = {'im': 'ax_flare.flc', 'x': 500, 'y': 700}
    SMAC = {'im': 'ax_unity.flc', 'x': 430, 'y': 370}
    SMACX = {'im': 'ax_alien.flc', 'x': 830, 'y': 430}


class WavFile(Enum):
    SMAC = 'fx/ax_unity.wav'
    SMACX = 'fx/ax_alien_shot.wav'


class SMACPack():
    def __init__(self):
        sdl2.SDL_Init(sdl2.SDL_INIT_VIDEO | sdl2.SDL_INIT_AUDIO)
        sdl2.sdlmixer.Mix_OpenAudio(22050, sdl2.sdlmixer.MIX_DEFAULT_FORMAT,
                                    1, 1024)
        centered = sdl2.SDL_WINDOWPOS_CENTERED
        self.window = sdl2.ext.Window(
                          'SMAC Planetary Pack',
                          size=(1024, 768),
                          position=(centered, centered),
                      )
        self.renderer = sdl2.ext.Renderer(self.window)

        self.font = FontTTF(data_filename('fonts/arialb.ttf'),
                            '15px', (255, 255, 255, 255))
        self.font.add_style('green', '15px', (84, 172, 148, 255))
        self.font.add_style('green_highlight', '15px', (176, 232, 188, 255))

        self.highlight = None

        self.sounds = {}
        for data in WavFile:
            self.sounds[data.name] = sdl2.sdlmixer.Mix_LoadWAV(
                bytes(data_filename(data.value), 'utf-8')
            )

        self.anims = {}
        ImageFile.LOAD_TRUNCATED_IMAGES = True
        for data in FlcFile:
            self.anims[data.name] = {
                'im': Image.open(data_filename(data.value['im'])),
                'count': 0,
            }

        self.background_image = self.init_background()

    def init_background(self) -> sdl2.surface.SDL_Surface:
        bg = sdl2.ext.load_img(data_filename('ax_background.pcx'))
        border = sdl2.ext.load_img(data_filename('axstart_border.pcx'))

        # Create a green version of 'bg'
        bg_green = sdl2.SDL_DuplicateSurface(bg)
        tmp = sdl2.SDL_DuplicateSurface(bg_green)
        sdl2.ext.fill(tmp, sdl2.ext.Color(0, 95, 20))
        sdl2.SDL_SetSurfaceAlphaMod(tmp, 100)
        sdl2.SDL_BlitSurface(tmp, None, bg_green, None)
        sdl2.SDL_FreeSurface(tmp)

        # Create a big black surface 'tmp' and center 'bg_green' on it
        tmp = sdl2.SDL_DuplicateSurface(border)
        sdl2.ext.fill(tmp, 0)
        rect = sdl2.rect.SDL_Rect(400, 300, 800, 600)
        sdl2.SDL_BlitSurface(bg_green, None, tmp, rect)

        # Create a big black surface 'final_bg' and center 'bg' on it
        final_bg = sdl2.SDL_DuplicateSurface(border)
        sdl2.ext.fill(final_bg, 0)
        rect = sdl2.rect.SDL_Rect(400, 300, 800, 600)
        sdl2.SDL_BlitSurface(bg, None, final_bg, rect)

        # Apply 'border' with AlphaColorKey to 'tmp' (black -> alpha)
        key = sdl2.pixels.SDL_MapRGB(border.format, 0, 0, 0)
        sdl2.SDL_SetColorKey(border, True, key)
        sdl2.SDL_BlitSurface(border, None, tmp, None)

        # Apply 'tmp' with AlphaColorKey to 'final_bg' (pink -> alpha)
        key = sdl2.pixels.SDL_MapRGB(border.format, 100, 16, 156)
        sdl2.SDL_SetColorKey(tmp, True, key)
        sdl2.SDL_BlitSurface(tmp, None, final_bg, None)

        sdl2.SDL_FreeSurface(bg)
        sdl2.SDL_FreeSurface(bg_green)
        sdl2.SDL_FreeSurface(border)
        sdl2.SDL_FreeSurface(tmp)

        return final_bg

    def show_splash_screen(self) -> None:
        splash = sdl2.ext.load_img(data_filename('loki_logo1024.pcx'))
        tx = sdl2.ext.Texture(self.renderer, splash)
        sdl2.SDL_FreeSurface(splash)
        self.renderer.copy(tx, dstrect=(0, 0))
        tx.destroy()
        self.renderer.present()
        sdl2.SDL_Delay(3000)

    def update_button(self) -> None:
        color_d = 'green'
        color_h = 'green_highlight'

        color = color_h if self.highlight == Button.SMAC else color_d
        txt_rendered = self.font.render_text('ALPHA CENTAURI',
                                             color, width=200)
        tx = sdl2.ext.Texture(self.renderer, txt_rendered)
        self.renderer.copy(tx, dstrect=(610, 448))
        tx.destroy()

        color = color_h if self.highlight == Button.SMACX else color_d
        txt_rendered = self.font.render_text('ALIEN CROSSFIRE',
                                             color, width=200)
        tx = sdl2.ext.Texture(self.renderer, txt_rendered)
        self.renderer.copy(tx, dstrect=(610, 523))
        tx.destroy()

        color = color_h if self.highlight == Button.CANCEL else color_d
        txt_rendered = self.font.render_text('CANCEL', color, width=200)
        tx = sdl2.ext.Texture(self.renderer, txt_rendered)
        self.renderer.copy(tx, dstrect=(610, 598))
        tx.destroy()

    def refresh_bg(
            self,
            bg: sdl2.surface.SDL_Surface,
            animations: list,
    ) -> None:
        for anim in animations:
            im = self.anims[anim]['im']
            self.anims[anim]['count'] += 1
            if self.anims[anim]['count'] >= im.n_frames:
                self.anims[anim]['count'] = 1

            im.seek(self.anims[anim]['count'])
            surface = sdl2.ext.image.pillow_to_surface(im)
            x = FlcFile[anim].value['x']
            y = FlcFile[anim].value['y']
            rect = sdl2.rect.SDL_Rect(x, y, im.width, im.height)
            sdl2.SDL_BlitSurface(surface, None, bg, rect)
            sdl2.SDL_FreeSurface(surface)

        tx = sdl2.ext.Texture(self.renderer, bg)
        self.renderer.copy(tx, srcrect=(288, 216, 1024, 768))
        tx.destroy()

        txt_rendered = self.font.render_text('CHOOSE YOUR WORLD', width=300)
        tx = sdl2.ext.Texture(self.renderer, txt_rendered)
        self.renderer.copy(tx, dstrect=(493, 101))
        tx.destroy()

        self.update_button()
        self.renderer.present()

    def launch(self, app: str) -> None:
        try:
            os.execvp(app, [app])
        except OSError as e:
            print("%s: %s" % (app, e))

    def run(self) -> int:
        rect_button_smac = sdl2.rect.SDL_Rect(575, 435, 295, 55)
        rect_button_smacx = sdl2.rect.SDL_Rect(575, 510, 295, 55)
        rect_button_quit = sdl2.rect.SDL_Rect(575, 585, 295, 55)
        rect_anim_smac = sdl2.rect.SDL_Rect(142, 154, 320, 256)
        rect_anim_smacx = sdl2.rect.SDL_Rect(542, 214, 274, 205)

        self.window.show()
        self.show_splash_screen()

        running = True
        animations = ['FLARE']
        while running:
            bg = sdl2.SDL_DuplicateSurface(self.background_image)

            events = sdl2.ext.get_events()
            for event in events:
                if event.type == sdl2.SDL_QUIT:
                    running = False
                    break
                elif event.type == sdl2.SDL_MOUSEBUTTONDOWN:
                    x = event.motion.x
                    y = event.motion.y
                    if is_in_rect(rect_button_smac, x, y):
                        self.launch('smac')
                    elif is_in_rect(rect_button_smacx, x, y):
                        self.launch('smacx')
                    elif is_in_rect(rect_button_quit, x, y):
                        running = False
                elif event.type == sdl2.SDL_MOUSEMOTION:
                    x = event.motion.x
                    y = event.motion.y
                    if is_in_rect(rect_button_smac, x, y) or \
                       is_in_rect(rect_anim_smac, x, y):
                        if self.highlight != Button.SMAC:
                            self.highlight = Button.SMAC
                            sdlmixer.Mix_PlayChannel(-1,
                                                     self.sounds['SMAC'], -1)
                            animations.append('SMAC')
                    elif is_in_rect(rect_button_smacx, x, y) or is_in_rect(
                                    rect_anim_smacx, x, y):
                        if self.highlight != Button.SMACX:
                            self.highlight = Button.SMACX
                            sdlmixer.Mix_PlayChannel(-1,
                                                     self.sounds['SMACX'], -1)
                            animations.append('SMACX')
                    elif is_in_rect(rect_button_quit, x, y):
                        if self.highlight != Button.CANCEL:
                            self.highlight = Button.CANCEL
                    else:
                        if self.highlight is not None:
                            if self.highlight != Button.CANCEL:
                                animations.remove(self.highlight.name)
                                self.anims[self.highlight.name]['count'] = 0
                            self.highlight = None
                            sdlmixer.Mix_HaltChannel(-1)

            self.refresh_bg(bg, animations)
            sdl2.SDL_FreeSurface(bg)
            sdl2.SDL_Delay(70)

        sdlmixer.Mix_CloseAudio()
        sdl2.ext.quit()
        return 0


if __name__ == "__main__":
    os.environ['SDL_VIDEO_WAYLAND_WMCLASS'] = 'smacpack'
    os.environ['SDL_VIDEO_X11_WMCLASS'] = 'smacpack'
    smackpack = SMACPack()
    sys.exit(smackpack.run())



#!/usr/bin/python3
# encoding=utf-8
#
# Copyright © 2018 Alexandre Detiste <alexandre@detiste.be>
# SPDX-License-Identifier: GPL-2.0-or-later

# Usage:
# curl -s -D headers.html \
# https://wiki.scummvm.org/index.php/Where_to_get_the_games > scummvm.html
# _build/run-tool-uninstalled tools/check_scummvm.py

import os
import re

from bs4 import BeautifulSoup

from game_data_packager.game import load_games

KEY = '22d200f8670dbdb3e253a90eee5098477c95c23d'


class Game:
    gog: str | None = None
    steam: int | None = None


def load_yaml() -> dict[str, Game]:
    print('# Reading data from YAML...')
    games = dict()
    for name, game in load_games(
        datadir=os.environ.get('GDP_BUILDDIR', 'out')
    ).items():
        game.load_file_data()
        if game.wikibase != 'https://wiki.scummvm.org/index.php/':
            continue
        assert game.wiki, name

        g = Game()
        for package in game.packages.values():
            gog = package.gog or game.gog
            steam = package.steam or game.steam
            if 'game' in gog:
                g.gog = gog['game']
            elif 'url' in gog:
                g.gog = gog['url']
            if 'id' in steam:
                g.steam = steam['id']

        games[game.wiki] = g
    return games


games = load_yaml()


# curl -s -D headers.html \
# https://wiki.scummvm.org/index.php/Where_to_get_the_games > scummvm.html


def load_html() -> dict[str, Game]:
    print('# Reading data from ScummVM wiki...')
    soup = BeautifulSoup(open("scummvm.html"), "html.parser")

    downloads: dict[str, Game] = dict()
    for table in soup.find_all('table', {'border': '1', 'width': '100%'}):
        print('## ' + table.find_previous('span').text)
        for row in table.find_all('tr'):
            game = None
            for link in row.find_all('a'):
                url = link['href']

                # URL that have been reviewed
                if url == 'https://www.scummvm.org/games/':
                    continue
                if url == 'https://www.msadams.com/downloads.htm':
                    continue
                if 'cdaccess.com/' in url:
                    continue
                if url.startswith('http://www.mdna-games.com'):
                    continue

                m = re.search(r'/index.php\?title\=(.*?)$', url)
                if m:
                    game = m.group(1)
                    downloads[game] = Game()
                    continue

                m = re.search(r'https://www.gog.com/game/(.*?)\?pp\=%s' % KEY,
                              url)
                if m:
                    if game is None:
                        print('Missing game id:', url)
                        continue
                    downloads[game].gog = m.group(1)
                    continue

                m = re.search(r'http://store.steampowered.com/app/(.*?)$', url)
                if m:
                    steam_id = int(m.group(1))
                    assert game is not None, url
                    downloads[game].steam = steam_id
                    continue

                print(game, url)
    return downloads


downloads = load_html()

print('# Join...')
no_game = Game()
for game, shop in sorted(games.items()):
    new_gog = downloads.get(game, no_game).gog
    if shop.gog != new_gog:
        print("%-70s" % game, shop.gog, new_gog)

    new_steam = downloads.get(game, no_game).steam
    if shop.steam != new_steam:
        print("%-70s" % game, shop.steam, new_steam)



#!/usr/bin/python3
#
# Copyright © 2023 Alexandre Detiste <alexandre@detiste.be>
# SPDX-License-Identifier: GPL-2.0-or-later

# match engines known to GDP versus all engines in contrib
# to check if we missed any -> we did missed openrct2

# TODO:
# - also read ITP/RFP from wnpp

import apt

from game_data_packager.game import (load_games)
from game_data_packager.packaging.deb import (DebPackaging)
packaging = DebPackaging()


def get_engines() -> set[str]:
    '''all engines known to GDP'''

    engines: set[str] = set()

    for name, game in load_games().items():
        game.load_file_data()
        if game.binary_executables:
            # GDP also ship the binary engine
            continue
        for package in game.packages.values():
            if package.section in ('fonts',):
                continue
            engine = package.engine or game.engine
            assert engine, name
            if type(engine) is dict:
                engine = packaging.substitute(engine, name)
            assert type(engine) is str
            engine = engine.split(' ')[0]
            engines.add(engine)
    return engines


engines = get_engines()


def get_tools() -> set[str]:
    '''Recommended & Suggests tools that may be in contrib'''
    tools: set[str] = set()
    return tools


cache = apt.Cache()

all: dict[str, str] = dict()
main: dict[str, str] = dict()
contrib: dict[str, str] = dict()
non_free: dict[str, str] = dict()

for k in cache.keys():
    p = cache[k]
    # print(dir(p))
    if p.candidate and p.candidate.downloadable:
        c = p.candidate
        # print(dir(c))
        if '/' in c.section:
            archive, section = c.section.split('/')
        else:
            archive = 'main'
            section = c.section
        all[p.shortname] = section
        if archive == 'contrib':
            contrib[p.shortname] = section
        elif archive == 'non-free':
            non_free[p.shortname] = section
        else:
            main[p.shortname] = section

for engine in sorted(engines):
    if engine not in all:
        print('%s is not packaged' % engine)
    elif all[engine] != 'games':
        print('%s is not a game: %s' % (engine, all[engine]))

for maybe_engine, section in sorted(contrib.items()):
    if maybe_engine in engines:
        pass
    elif section == 'games':
        print(maybe_engine)

for maybe_engine, section in sorted(non_free.items()):
    if maybe_engine in engines:
        pass
    elif section == 'games':
        print(maybe_engine)



#!/usr/bin/python3
# encoding=utf-8
#
# Copyright © 2016 Simon McVittie <smcv@debian.org>
# SPDX-License-Identifier: GPL-2.0-or-later

from __future__ import annotations

import logging
import os
from collections.abc import (Sequence)
from typing import (TYPE_CHECKING)

from ..build import (PackagingTask)
from ..game import (GameData)
from ..util import (TemporaryUmask, mkdir_p)

if TYPE_CHECKING:
    from typing import (Unpack)
    from ..data import (Package)
    from ..packaging import (PerPackageState, PackagingTaskArgs)

logger = logging.getLogger(__name__)


class UnrealTask(PackagingTask):
    def fill_extra_files(
        self,
        per_package_state: PerPackageState
    ) -> None:
        super().fill_extra_files(per_package_state)
        package = per_package_state.package
        destdir = per_package_state.destdir

        if package.name == 'unreal-data':
            with TemporaryUmask(0o022):
                self.__convert_logo(destdir, package, 'skaarj_logo.jpg')
        elif package.name == 'ut99-data':
            with TemporaryUmask(0o022):
                self.__convert_logo(destdir, package, 'ut99.png')
        elif package.name == 'ut2004-data':
            with TemporaryUmask(0o022):
                self.__convert_logo(destdir, package, 'ut2004.png')

        if package.name in ('unreal-gold', 'unreal-classic', 'ut99-data'):
            with TemporaryUmask(0o022):
                self.__add_manifest(package, destdir)

    def __add_manifest(self, package: Package, destdir: str) -> None:
        # A real Manifest.ini is much larger than this, but this is
        # enough to identify the version.

        install_to = self.packaging.substitute(
            package.install_to,
            package.name,
        )
        system = os.path.join(destdir, install_to.strip('/'), 'System')
        assert system.startswith(destdir + '/'), (system, destdir)
        mkdir_p(system)

        with open(os.path.join(system, 'Manifest.ini'), 'w') as writer:
            if package.name == 'unreal-gold':
                groups: Sequence[tuple[str, str, str]] = (
                    ('UnrealGold', package.name, package.version),
                    ('Unreal Gold', package.name, package.version),
                )
                sample_file = 'System\\UnrealLinux.ini'
            elif package.name == 'unreal-classic':
                groups = (('Unreal', package.name, package.version),)
                sample_file = 'System\\UnrealLinux.ini'
            elif package.name == 'ut99-data':
                # The unofficial patches after 436 do not update the manifest
                groups = (('UnrealTournament', 'ut99', '436'),)
                sample_file = 'System\\UnrealTournament.ini'
            else:
                raise AssertionError(
                    'Method should not be called for this package',
                )

            lines = ['[Setup]', 'MasterProduct=' + groups[0][0]]

            for g in groups:
                lines.append('Group=' + g[0])

            for g in groups:
                lines.append('')
                lines.append('[' + g[0] + ']')
                lines.append('Caption=' + g[1])
                lines.append('Version=' + g[2])
                lines.append('File=' + sample_file)

            lines += [
                '',
                '[RefCounts]',
                'File:%s=%d' % (sample_file, len(groups)),
                '',
            ]

            for line in lines:
                writer.write(line + '\r\n')

    def __convert_logo(
        self,
        destdir: str,
        package: Package,
        logo_name: str,
    ) -> None:
        source_logo = os.path.join(
            destdir,
            self.packaging.substitute(
                package.install_to, package.name,
            ).strip('/'),
            logo_name,
        )
        assert source_logo.startswith(destdir + '/'), (source_logo, destdir)

        try:
            import gi
            gi.require_version('GdkPixbuf', '2.0')
            from gi.repository import GdkPixbuf
        except Exception:
            logger.warning(
                'Unable to load GdkPixbuf bindings. %s icon '
                'will not be resized or converted',
                self.game.longname,
            )
            return

        mkdir_p(os.path.join(destdir, 'usr', 'share', 'icons',
                'hicolor', '48x48', 'apps'))
        mkdir_p(os.path.join(destdir, 'usr', 'share', 'icons',
                'hicolor', '256x256', 'apps'))

        pixbuf = GdkPixbuf.Pixbuf.new_from_file(source_logo)
        assert pixbuf
        pixbuf.savev(
            os.path.join(
                destdir, 'usr', 'share', 'icons',
                'hicolor', '256x256', 'apps', self.game.shortname + '.png'
            ),
            'png', [], [],
        )

        pixbuf = GdkPixbuf.Pixbuf.new_from_file_at_size(source_logo, 48, 48)
        assert pixbuf
        pixbuf.savev(
            os.path.join(
                destdir, 'usr', 'share', 'icons',
                'hicolor', '48x48', 'apps', self.game.shortname + '.png',
            ),
            'png', [], [],
        )


class UnrealGameData(GameData):
    def construct_task(
        self,
        **kwargs: Unpack[PackagingTaskArgs]
    ) -> UnrealTask:
        return UnrealTask(self, **kwargs)


GAME_DATA_SUBCLASS = UnrealGameData



#!/usr/bin/python3
# encoding=utf-8
#
# Copyright © 2015 Alexandre Detiste <alexandre@detiste.be>
# SPDX-License-Identifier: GPL-2.0-or-later

# a simple spider to locate Wikipedia url
# in per-engine-wiki pages
# we don't rescan games we already have

import sys
import time
import urllib.request

from bs4 import BeautifulSoup

from game_data_packager.game import load_games
from game_data_packager.util import AGENT

denylist = ['compet-n', 'hacx']

try:
    todo = sys.argv[1]
except IndexError:
    todo = '*'


def is_wikipedia(href: str) -> bool:
    return bool(href) and "wikipedia" in href


def add_wikipedia_to_game_yaml(game: str, url: str) -> None:
    print('Add %s to game %s' % (url, game))
    filename = f"data/{game}.yaml"
    with open(filename) as fh:
        data = fh.read()

    tags = ['wiki', 'genre', 'copyright', 'longname']
    idx = 0
    for tag in tags:
        substring = f"\n{tag}: "
        if substring in data:
            idx1 = data.index(substring) + 1
            idx2 = data[idx1:].index('\n') + 1
            idx = idx1 + idx2
            break

    data = data[:idx] + f"wikipedia: {url}\n" + data[idx:]
    with open(filename, "w") as fh:
        fh.write(data)


for shortname, game in load_games(game=todo, use_vfs=False).items():
    if game.wikipedia is not None:
        continue
    if shortname in denylist:
        continue
    if not game.wiki:
        continue

    print('processing %s ...' % shortname)
    url = game.wikibase + game.wiki
    try:
        html = urllib.request.urlopen(urllib.request.Request(url,
                                      headers={'User-Agent': AGENT}))
    except urllib.error.HTTPError as e:
        print('%s : %s' % (url, e))
        continue

    soup = BeautifulSoup(html, 'lxml')
    for tag in soup.find_all(href=is_wikipedia):
        print('  ' + tag['href'])
        add_wikipedia_to_game_yaml(shortname, tag['href'])

    # break
    time.sleep(1)



#!/usr/bin/python3
# encoding=utf-8
#
# Copyright © 2023 Sébastien Noel <sebastien@twolife.be>
# SPDX-License-Identifier: GPL-2.0-or-later

from __future__ import annotations

import logging
import os
import shutil
import subprocess
from typing import (TYPE_CHECKING)

from ..build import (PackagingTask)
from ..game import (GameData)

if TYPE_CHECKING:
    from typing import (Unpack)
    from ..packaging import (PerPackageState, PackagingTaskArgs)

logger = logging.getLogger(__name__)


class LokiShimGameData(GameData):
    def construct_task(self, **kwargs: Unpack[PackagingTaskArgs]) -> LokiShim:
        return LokiShim(self, **kwargs)


class LokiShim(PackagingTask):
    def fill_extra_files(
        self,
        per_package_state: PerPackageState
    ) -> None:
        super().fill_extra_files(per_package_state)
        package = per_package_state.package
        destdir = per_package_state.destdir

        install_dir = os.path.join(destdir, 'usr', 'lib', self.game.shortname)

        if package.name == 'civctp-bin':
            fixed_bin = os.path.join(
                            self.get_workdir(), 'civctp.dynamic_fixed'
                        )
            shutil.copyfile(
                os.path.join(install_dir, 'civctp.dynamic'),
                fixed_bin
            )

            # https://davidgow.net/hacks/civctp.html
            subprocess.check_call([
                'patchelf', '--replace-needed',
                'libSDL_mixer-1.0.so.0', 'libSDL_mixer-1.2.so.0', fixed_bin
            ])
            subprocess.check_call([
                'patchelf', '--replace-needed',
                'libSDL-1.1.so.0', 'libSDL-1.2.so.0', fixed_bin
            ])

            regex = 's/' \
                    r'\x{E8}\x{F5}\x{96}\x{B6}\x{FF}\x{89}\x{C0}\x{89}' \
                    r'\x{45}\x{FC}\x{8B}\x{43}\x{04}\x{3B}\x{45}\x{FC}' \
                    r'\x{75}\x{05}\x{FF}\x{43}\x{08}\x{EB}\x{14}/' \
                    r'\x{90}\x{90}\x{90}\x{90}\x{90}\x{90}\x{90}\x{90}' \
                    r'\x{90}\x{90}\x{90}\x{90}\x{90}\x{90}\x{90}\x{90}' \
                    r'\x{90}\x{90}\x{90}\x{90}\x{90}\x{90}\x{90}/'
            subprocess.check_call([
                'perl', '-pi', '-wE', regex, fixed_bin
            ])

            regex = 's/' \
                    r'\x{83}\x{7B}\x{08}\x{00}\x{7E}\x{05}\x{FF}\x{4B}' \
                    r'\x{08}\x{EB}\x{15}/' \
                    r'\x{90}\x{90}\x{90}\x{90}\x{90}\x{90}\x{90}\x{90}' \
                    r'\x{90}\x{90}\x{90}/'
            subprocess.check_call([
                'perl', '-pi', '-wE', regex, fixed_bin
            ])

            subprocess.check_call([
                'install', '-s', '-m755', fixed_bin,
                os.path.join(install_dir, 'civctp.dynamic_fixed'),
            ])


GAME_DATA_SUBCLASS = LokiShimGameData



#!/usr/bin/python3
# encoding=utf-8
#
# Copyright © 2016 Alexandre Detiste <alexandre@detiste.be>
# Copyright © 2016 Simon McVittie <smcv@debian.org>
# SPDX-License-Identifier: GPL-2.0-or-later

from __future__ import annotations

import logging
import os
from typing import (TYPE_CHECKING)

from ..build import (PackagingTask)
from ..game import (GameData)
from ..util import (check_call)

if TYPE_CHECKING:
    from typing import (Unpack)
    from ..packaging import (PerPackageState, PackagingTaskArgs)

logger = logging.getLogger(__name__)


class MorrowindGameData(GameData):
    def construct_task(
        self,
        **kwargs: Unpack[PackagingTaskArgs]
    ) -> MorrowindTask:
        return MorrowindTask(self, **kwargs)


class MorrowindTask(PackagingTask):
    def fill_extra_files(
        self,
        per_package_state: PerPackageState
    ) -> None:
        super().fill_extra_files(per_package_state)
        package = per_package_state.package
        destdir = per_package_state.destdir

        if package.expansion_for:
            return

        install_to = self.packaging.substitute(package.install_to,
                                               package.name)

        datadir = os.path.join(destdir, install_to.strip('/'))
        assert datadir.startswith(destdir + '/'), (datadir, destdir)

        subdir_expected_by_iniimporter = os.path.join(datadir, 'Data Files')
        os.symlink(datadir, subdir_expected_by_iniimporter)

        ini = os.path.join(datadir, 'Morrowind.ini')

        if os.path.exists(ini + '.gog'):
            # The GOG ini file already contains Tribunal and Bloodmoon
            initext = open(ini + '.gog', encoding='latin-1').read()
            open(ini, 'w', encoding='latin-1').write(initext)

        elif os.path.exists(ini + '.orig'):
            initext = open(ini + '.orig', encoding='latin-1').read()

            if package.name.startswith(     # either of:
                ('morrowind-tribunal', 'morrowind-complete'),
            ):
                initext = initext + '\r\nGameFile1=Tribunal.esm\r\n'

            if package.name.startswith('morrowind-bloodmoon'):
                initext = initext + '\r\nGameFile1=Bloodmoon.esm\r\n'

            if package.name.startswith('morrowind-complete'):
                initext = initext + '\r\nGameFile2=Bloodmoon.esm\r\n'

            open(ini, 'w', encoding='latin-1').write(initext)

        cfg = os.path.join(datadir, 'openmw.cfg')

        check_call(['openmw-iniimporter',
                    '--verbose',
                    '--game-files',
                    '--encoding', 'win1252',
                    '--ini', ini,
                    '--cfg', cfg])
        os.unlink(subdir_expected_by_iniimporter)

        with open(cfg, 'a', encoding='utf-8') as f:
            f.write('data=%s\n' % os.path.join('/', install_to))

        # then user needs to do this:
        #
        # $ mkdir -p ~/.config/openmw/
        # $ cp /usr/share/games/morrowind-fr/openmw.cfg ~/.config/openmw/


# See morrowind.txt for additional notes


GAME_DATA_SUBCLASS = MorrowindGameData



#!/usr/bin/python3
# encoding=utf-8
#
# Copyright © 2014-2016 Simon McVittie <smcv@debian.org>
# Copyright © 2015-2016 Alexandre Detiste <alexandre@detiste.be>
# SPDX-License-Identifier: GPL-2.0-or-later

from __future__ import annotations

import argparse
import logging
import os
import sys
import textwrap
import time
import zipfile
from types import TracebackType
from typing import (Any, NoReturn)

from .config import (read_config)
from .data import (ProgressCallback)
from .game import (load_games)
from .packaging import (get_packaging_system)
from .paths import (DATADIR)
from .util import (human_size)
from .version import (FORMAT, DISTRO, GAME_PACKAGE_VERSION)

logger = logging.getLogger(__name__)


class TerminalProgress(ProgressCallback):
    def __init__(
        self,
        info: str | None = None,
        *,
        interval: float = 0.2,
    ) -> None:
        """Constructor.

        Progress will update at most once per @interval seconds, unless we
        are at a "checkpoint".
        """
        self.pad = ' '
        self.interval = interval
        self.ts = time.time()
        self.info = info

    def __call__(
        self,
        done: int | None = None,
        total: int | None = None,
        checkpoint: bool = False
    ) -> None:
        ts = time.time()

        if done is None or (ts < self.ts + self.interval and not checkpoint):
            return

        if self.info is not None:
            print(self.info, file=sys.stderr)
            self.info = None

        if done is None or total == 0:
            s = ''
        elif total is None or total == 0:
            s = human_size(done)
        else:
            s = '%.0f%% %s/%s' % (
                100 * done / total,
                human_size(done),
                human_size(total),
            )

        self.ts = ts

        if len(self.pad) <= len(s):
            self.pad = ' ' * len(s)

        print(' %s \r %s\r' % (self.pad, s), end='', file=sys.stderr)

    def __enter__(self) -> TerminalProgress:
        return self

    def __exit__(
        self,
        _et: type[BaseException] | None = None,
        _ev: BaseException | None = None,
        _tb: TracebackType | None = None
    ) -> None:
        self()


def run_command_line() -> None:
    logger.debug('Arguments: %r', sys.argv)

    # Don't set any defaults on these base parsers, because that interferes
    # with the ability to recognise the same argument either before or
    # after the game name. Set them on the Namespace instead.
    base_parser = argparse.ArgumentParser(
        prog='game-data-packager',
        description='Package game files.',
        add_help=False,
        argument_default=argparse.SUPPRESS,
    )

    class VersionAction(argparse.Action):
        def __call__(
            self,
            parser: Any,
            namespace: argparse.Namespace,
            values: Any,
            option_string: str | None = None
        ) -> None:
            print(GAME_PACKAGE_VERSION)
            exit(0)

    base_parser.add_argument(
        '--version', action=VersionAction, nargs=0,
        help='display version and exit',
    )

    group = base_parser.add_mutually_exclusive_group()
    group.add_argument(
        '--verbose', action='store_true',
        help='show output from external tools',
    )
    group.add_argument(
        '--no-verbose', action='store_false',
        dest='verbose',
        help='hide output from external tools (default)',
    )

    class DebugAction(argparse.Action):
        def __call__(
            self,
            parser: Any,
            namespace: argparse.Namespace,
            values: Any,
            option_string: str | None = None
        ) -> None:
            logging.getLogger().setLevel(logging.DEBUG)

    base_parser.add_argument(
        '--debug', action=DebugAction, nargs=0,
        help='show debug messages',
    )

    game_parser = argparse.ArgumentParser(
        prog='game-data-packager',
        description='Package game files.',
        add_help=False,
        argument_default=argparse.SUPPRESS,
        parents=(base_parser,),
    )

    game_parser.add_argument(
        '--everything', action='store_true',
        help='Download all possible expansions',
    )
    game_parser.add_argument(
        '--package', '-p', action='append',
        dest='packages', metavar='PACKAGE',
        help='Produce this data package (may be repeated)',
    )

    game_parser.add_argument(
        '--target-format',
        choices='arch deb rpm'.split(),
        help='Produce packages for this packaging system',
    )
    game_parser.add_argument(
        '--target-distro',
        help='Produce packages suitable for this distro',
    )
    game_parser.add_argument(
        '--target-architecture',
        help=(
            'Produce packages suitable for this architecture (named '
            'according to either the --target-distro or dpkg conventions)'
        ),
    )

    game_parser.add_argument(
        '--install-method', metavar='METHOD',
        dest='install_method',
        help=(
            'Use METHOD (apt, dpkg, gdebi, gdebi-gtk, gdebi-kde) '
            'to install packages'
        ),
    )

    game_parser.add_argument(
        '--gain-root-command', metavar='METHOD',
        dest='gain_root_command',
        help='Use METHOD (su, sudo, pkexec) to gain root if needed',
    )

    # This is about ability to audit, not freeness. We don't have an
    # option to restrict g-d-p to dealing with Free Software, because
    # that would rule out the vast majority of its packages: if a game's
    # data is Free Software, we could put it in main or contrib and not need
    # g-d-p at all.
    game_parser.add_argument(
        '--binary-executables', action='store_true',
        help=(
            'allow installation of executable code that was not built '
            'from public source code'
        ),
    )
    game_parser.add_argument(
        '--demo', action='store_true',
        help=(
            'Build demo/shareware packages even if files for a full version '
            'are available'
        ),
    )

    # Misc options
    group = game_parser.add_mutually_exclusive_group()
    group.add_argument(
        '-i', '--install', action='store_true',
        help='install the generated package',
    )
    group.add_argument(
        '--force-install', action='store_true',
        help='install the generated package without prompting',
    )
    group.add_argument(
        '-n', '--no-install', action='store_false',
        dest='install',
        help='do not install the generated package (requires -d, default)',
    )

    game_parser.add_argument(
        '-d', '--destination', metavar='OUTDIR',
        help='write the generated .%s(s) to OUTDIR' % FORMAT,
    )

    group = game_parser.add_mutually_exclusive_group()
    group.add_argument(
        '-z', '--compress', action='store_true',
        help='compress generated .%s (default if -d is used)' % FORMAT,
    )
    group.add_argument(
        '--no-compress', action='store_false', dest='compress',
        help='do not compress generated .%s (default without -d)' % FORMAT,
    )

    group = game_parser.add_mutually_exclusive_group()
    group.add_argument(
        '--download', action='store_true',
        help='automatically download necessary files if possible (default)',
    )
    group.add_argument(
        '--no-download', action='store_false', dest='download',
        help='do not download anything',
    )
    game_parser.add_argument(
        '--save-downloads', metavar='DIR',
        help='save downloaded files to DIR, and look for files there',
    )

    group = game_parser.add_mutually_exclusive_group()
    group.add_argument(
        '--search', action='store_true',
        help=(
            'look for installed files in Steam and other likely places '
            '(default)'
        ),
    )
    group.add_argument(
        '--no-search', action='store_false', dest='search',
        help='only look in paths provided on the command line',
    )

    class DumbParser(argparse.ArgumentParser):
        def error(self, message: Any) -> NoReturn:  # type: ignore[empty-body]
            pass

    dumb_parser = DumbParser(parents=(game_parser,), add_help=False)
    dumb_parser.add_argument('game', type=str, nargs='?')
    dumb_parser.add_argument('paths', type=str, nargs='*')
    dumb_parser.add_argument('-h', '--help', action='store_true', dest='h')
    try:
        g = dumb_parser.parse_args().game
    except TypeError:
        g = None
    zip = os.path.join(DATADIR, 'vfs.zip')
    if g is None:
        games = load_games()
    elif '-h' in sys.argv or '--help' in sys.argv:
        games = load_games()
    elif os.path.isfile(os.path.join(DATADIR, '%s.json' % g)):
        games = load_games(game=g)
    elif not os.path.isfile(zip):
        games = load_games()
    else:
        with zipfile.ZipFile(zip, 'r') as zf:
            if '%s.json' % g in zf.namelist():
                games = load_games(game=g)
            else:
                games = load_games()

    parser = argparse.ArgumentParser(
        prog='game-data-packager',
        description='Package game files.', parents=(game_parser,),
        epilog=(
            'Run "game-data-packager GAME --help" to see game-specific '
            'arguments.'
        ),
    )

    game_parsers = parser.add_subparsers(
        dest='shortname',
        title='supported games',
        metavar='GAME',
    )

    for g in sorted(games.keys()):
        games[g].add_parser(game_parsers, game_parser)

    # GOG meta-mode
    gog_parser = game_parsers.add_parser(
        'gog',
        help='Package all your GOG.com games at once',
        description='Automatically package all your GOG.com games',
        formatter_class=argparse.RawDescriptionHelpFormatter,
        parents=(game_parser,),
    )
    group = gog_parser.add_mutually_exclusive_group()
    group.add_argument('--all', action='store_true', default=False,
                       help='show all GOG.com games')
    group.add_argument('--new', action='store_true', default=False,
                       help='show all new GOG.com games')

    # Steam meta-mode
    steam_parser = game_parsers.add_parser(
        'steam',
        help='Package all Steam games at once',
        description='Automatically package all your Steam games',
        formatter_class=argparse.RawDescriptionHelpFormatter,
        parents=(game_parser,),
    )
    group = steam_parser.add_mutually_exclusive_group()
    group.add_argument('--all', action='store_true', default=False,
                       help='package all Steam games')
    group.add_argument('--new', action='store_true', default=False,
                       help='package all new Steam games')

    mt_parser = game_parsers.add_parser(
        'make-template',
        help='Capture details of a new game or update an existing one',
        description=textwrap.dedent('''\
        This command collects details of game files in a directory or archive,
        so that they can be used for a newly supported game or added to an
        existing game. Please send its output to the Debian bug tracking system
        as a bug with 'wishlist' severity.\
        '''),
        formatter_class=argparse.RawDescriptionHelpFormatter,
        parents=(base_parser,),
    )
    mt_parser.add_argument('args', nargs='*', metavar='DEB|DIRECTORY|FILE')
    mt_parser.add_argument(
        '--strip-path', action='append', dest='strip_paths', default=[],
        help='Strip a prefix from all filenames (may be repeated)')
    mt_parser.add_argument(
        '--base', default=None, help='Base the template on an existing game')
    mt_parser.add_argument(
        '--template', action='append', dest='templates', default=[],
        help='Load pre-generated templates')
    mt_parser.add_argument(
        '-l', '--lower', action='store_true', dest='lower',
        help='make all files lowercase')
    mt_parser.add_argument(
        '-e', '--execute', action='store_true', dest='execute',
        help='run this game through strace and see which files from '
             '/usr/share/games or /usr/local/games are needed')
    mt_parser.add_argument(
        '-f', '--flacsums', action='store_true', dest='flacsums',
        help='compute "flacsums" from .wav files')

    config = read_config()
    parsed = argparse.Namespace(
            binary_executables=False,
            compress=None,
            demo=False,
            destination=None,
            download=True,
            everything=False,
            force_install=False,
            verbose=False,
            install=False,
            install_method='',
            gain_root_command='',
            packages=[],
            save_downloads=None,
            search=True,
            shortname=None,
            target_architecture=None,
            target_format=FORMAT,
            target_distro=DISTRO,
    )
    if config['install']:
        logger.debug('obeying INSTALL=yes in configuration')
        parsed.install = True
    if config['preserve']:
        logger.debug('obeying PRESERVE=yes in configuration')
        parsed.destination = '.'
    if config['verbose']:
        logger.debug('obeying VERBOSE=yes in configuration')
        parsed.verbose = True
    if config['install_method']:
        logger.debug(
            'obeying INSTALL_METHOD=%r in configuration',
            config['install_method'],
        )
        parsed.install_method = config['install_method']
    if config['gain_root_command']:
        logger.debug(
            'obeying GAIN_ROOT_COMMAND=%r in configuration',
            config['gain_root_command'],
        )
        parsed.gain_root_command = config['gain_root_command']

    parser.parse_args(namespace=parsed)
    parsed.install = parsed.install or parsed.force_install
    logger.debug('parsed command-line arguments into: %r', parsed)

    if (parsed.destination is None and not parsed.install and
            parsed.shortname != 'make-template'):
        logger.error('At least one of --install or --destination is required')
        sys.exit(2)

    if parsed.shortname is None:
        parser.print_help()
        sys.exit(0)

    for arg, path in (('--save-downloads', parsed.save_downloads),
                      ('--destination', parsed.destination)):
        if path is None:
            continue
        elif not os.path.isdir(path):
            logger.error('argument "%s" to %s does not exist', path, arg)
            sys.exit(2)
        elif not os.access(path, os.W_OK | os.X_OK):
            logger.error('argument "%s" to %s is not writable', path, arg)
            sys.exit(2)
        elif path == '.':
            try:
                os.getcwd()
            except FileNotFoundError:
                logger.error('argument "%s" to %s has been deleted', path, arg)
                sys.exit(2)

    if parsed.shortname == 'steam':
        from .steam import (run_steam_meta_mode)
        run_steam_meta_mode(parsed, games)
        return
    elif parsed.shortname == 'gog':
        from .gog import (run_gog_meta_mode)
        run_gog_meta_mode(parsed, games)
        return
    elif parsed.shortname == 'make-template':
        import game_data_packager.make_template
        game_data_packager.make_template.main(parsed, games)
        return
    elif parsed.shortname in games:
        game = games[parsed.shortname]
    else:
        parsed.package = parsed.shortname
        for game in games.values():
            if parsed.shortname in game.packages:
                break
            if parsed.shortname in game.aliases:
                break
        else:
            raise AssertionError('could not find %s' % parsed.shortname)

    with game.construct_task(
        packaging=get_packaging_system(
            parsed.target_format,
            parsed.target_distro,
            parsed.target_architecture,
        ),
    ) as task:
        if sys.stderr.isatty():
            task.progress_factory = TerminalProgress

        task.run_command_line(parsed)


if __name__ == '__main__':
    run_command_line()
